import BiomationScripter as _BMS
import BiomationScripter.OTProto as _OTProto
from opentrons import simulate as OT2
import math
import warnings
from typing import List, NewType, Tuple, Union, Literal
from collections import namedtuple

class Template(_OTProto.OTProto_Template):
    def __init__(self,
        DNA: List[str],
        DNA_Source_Type: str,
        DNA_Source_Wells: List[str],
        Primers: List[str],
        Primer_Source_Wells: List[str],
        Primer_Source_Type: str,
        DNA_Primer_Mixtures: List[ List[str] ],
        Destination_Type: str,
        DNA_Volume: float,
        Primer_Volume: float,
        Final_Volume: float,
        Water_Source_Type: Union[str, None],
        Water_Aliquot_Volume: float,
        DNA_Primers_Same_Source: bool = False,
        **kwargs
    ):

        ########################################
        # User defined aspects of the protocol #
        ########################################
        self.dna_primer_mixtures = DNA_Primer_Mixtures
        self.dna_volume = DNA_Volume
        self.primer_volume = Primer_Volume
        self.final_volume = Final_Volume

        ####################
        # Source materials #
        ####################
        self.dna = DNA
        self.primers = Primers

        self.dna_source_wells = DNA_Source_Wells
        self.primer_source_wells = Primer_Source_Wells

        self.dna_source_type = DNA_Source_Type
        self.primer_source_type = Primer_Source_Type
        self.water_source_type = Water_Source_Type

        self.water_aliquot_volume = Water_Aliquot_Volume

        self.dna_primers_same_source = DNA_Primers_Same_Source

        #######################
        # Destination Labware #
        #######################
        self.destination_type = Destination_Type

        ##################################################
        # Protocol Metadata and Instrument Configuration #
        ##################################################
        super().__init__(**kwargs)

    def run(self):

        #################
        # Load pipettes #
        #################
        self.load_pipettes()

        ################################
        # Calculate and load tip boxes #
        ################################
        # Number of tips needed is:
        # > One tip for the water total - volume = final - (DNA + Primer)
        # > One tip per DNA sample per destination
        # > One tip per primer per destination

        # Calculate and store all of the transfers which need to be performed by this protocol
        Water_Transfers = []
        DNA_Transfers = []
        Primer_Transfers = []
        for mixture in self.dna_primer_mixtures:
            dna = mixture[0]
            primers = mixture[1:]
            water_vol = self.final_volume - (self.dna_volume + len(primers) * self.primer_volume)
            DNA_Transfers.append(self.dna_volume)
            for primer in primers:
                Primer_Transfers.append(self.primer_volume)
            Water_Transfers.append(water_vol)

        # This calculates the number of each tip type needed, and adds them to the tips_needed attribute
        water_tips_needed = _OTProto.calculate_tips_needed(self._protocol, Water_Transfers, template = self, new_tip = False)
        dna_tips_needed = _OTProto.calculate_tips_needed(self._protocol, DNA_Transfers, template = self, new_tip = True)
        primer_tips_needed = _OTProto.calculate_tips_needed(self._protocol, Primer_Transfers, template = self, new_tip = True)

        # Add required number of tip boxes to the loaded pipettes
        self.add_tip_boxes_to_pipettes()

        #######################
        # Load source labware #
        #######################
        # Calculate how many water aliquots are required
        if self.water_aliquot_volume == 0:
            number_water_aliquots = 0
        else:
            number_water_aliquots = math.ceil(sum(Water_Transfers) / (self.water_aliquot_volume - self.water_aliquot_volume * 0.01))

            # Load the required number of water labware
            Water_Labware, Water_Locations = _OTProto.calculate_and_load_labware(
                                                        protocol = self._protocol,
                                                        labware_api_name = self.water_source_type,
                                                        wells_required = number_water_aliquots,
                                                        custom_labware_dir = self.custom_labware_dir
            )

        #########################################
        # Create labware layout for DNA labware #
        #########################################

        dna_layout = _BMS.Labware_Layout(
                                Name = "DNA",
                                Type = self.dna_source_type
        )
        n_dna_layout_rows, n_dna_layout_cols = _OTProto.get_labware_format(self.dna_source_type, self.custom_labware_dir)
        dna_layout.define_format(n_dna_layout_rows, n_dna_layout_cols)

        # Add content to the layout
        for dna, dna_well in zip(self.dna, self.dna_source_wells):
         dna_layout.add_content(
                        Well = dna_well,
                        Reagent = dna,
                        Volume = 30 # This refers to nothing important - should add in a calculator somewhere to let user know how much is needed
         )

        ###########################################################################################
        # Create labware layout for Primer labware (if primer and dna aren't in the same labware) #
        ###########################################################################################

        if not self.dna_primers_same_source:
            primer_layout = _BMS.Labware_Layout(
                                       Name = "Primers",
                                       Type = self.primer_source_type
            )
            n_primer_layout_rows, n_primer_layout_cols = _OTProto.get_labware_format(self.primer_source_type, self.custom_labware_dir)
            primer_layout.define_format(n_primer_layout_rows, n_primer_layout_cols)
        else:
            dna_layout.name = "DNA and Primers"
            primer_layout = dna_layout

        # Add content to the layout
        for primer, primer_well in zip(self.primers, self.primer_source_wells):
            primer_layout.add_content(
                           Well = primer_well,
                           Reagent = primer,
                           Volume = 30 # This refers to nothing important - should add in a calculator somewhere to let user know how much is needed
            )

        ###################################
        # Load labware from layout object #
        ###################################
        dna_labware = _OTProto.load_labware_from_layout(
                                Protocol = self._protocol,
                                Plate_Layout = dna_layout,
                                custom_labware_dir = self.custom_labware_dir
        )

        if not self.dna_primers_same_source:
            primer_labware = _OTProto.load_labware_from_layout(
                                    Protocol = self._protocol,
                                    Plate_Layout = primer_layout,
                                    custom_labware_dir = self.custom_labware_dir
            )
        else:
            primer_labware = dna_labware

        ############################
        # Load destination labware #
        ############################
        destination_labware, destination_locations = _OTProto.calculate_and_load_labware(
                                                       protocol = self._protocol,
                                                       labware_api_name = self.destination_type,
                                                       wells_required = len(self.dna_primer_mixtures),
                                                       custom_labware_dir = self.custom_labware_dir
        )

        ##################################
        # Start of protocol instructions #
        ##################################

        # User prompts to ensure deck is correctly set up
        if not _OTProto.get_p20 == None:
            self._protocol.pause("This protocol uses {} 20 uL tip boxes".format(_OTProto.tip_racks_needed(self.tips_needed["p20"], self.starting_tips["p20"])))
        if not _OTProto.get_p300 == None:
            self._protocol.pause("This protocol uses {} 300 uL tip boxes".format(_OTProto.tip_racks_needed(self.tips_needed["p300"], self.starting_tips["p300"])))
        if not _OTProto.get_p1000 == None:
            self._protocol.pause("This protocol uses {} 1000 uL tip boxes".format(_OTProto.tip_racks_needed(self.tips_needed["p1000"], self.starting_tips["p1000"])))

        for dna, dna_well in zip(self.dna, self.dna_source_wells):
            self._protocol.pause("Ensure DNA {} is loaded in slot {} of labware {} on slot {}".format(dna, dna_well, dna_labware.name, dna_labware.parent))

        for primer, primer_well in zip(self.primers, self.primer_source_wells):
            self._protocol.pause("Ensure Primer {} is loaded in slot {} of labware {} on slot {}".format(primer, primer_well, primer_labware.name, primer_labware.parent))

        if number_water_aliquots > 0:
            self._protocol.pause("This protocol uses {} aliquots of {} uL water, located at {}".format(len(Water_Locations), self.water_aliquot_volume, Water_Locations))

            ######################################
            # Transfer water into required wells #
            ######################################
            _OTProto.dispense_from_aliquots(
                                Protocol = self._protocol,
                                Transfer_Volumes = Water_Transfers,
                                Aliquot_Source_Locations = Water_Locations,
                                Destinations = destination_locations,
                                Aliquot_Volumes = self.water_aliquot_volume,
                                new_tip = False
            )

        ####################################
        # Transfer DNA into required wells #
        ####################################

        ## Get transfer locations for each of the DNA samples required
        dna_wells = []
        for mixture in self.dna_primer_mixtures:
            dna_name = mixture[0]
            dna_wells.append(dna_layout.get_wells_containing_liquid(dna_name)[0])

        DNA_Transfer_Source_Locations = _OTProto.get_locations(
                                                Labware = dna_labware,
                                                Wells = dna_wells
        )

        _OTProto.transfer_liquids(
                            Protocol = self._protocol,
                            Transfer_Volumes = DNA_Transfers,
                            Source_Locations = DNA_Transfer_Source_Locations,
                            Destination_Locations = destination_locations,
                            new_tip = True,
                            mix_before = (3, "transfer_volume"),
                            mix_after = (3, "transfer_volume")
        )

        ########################################
        # Transfer primers into required wells #
        ########################################

        ## Get transfer locations for each of the primers required
        primer_wells = []
        primer_destination_locations = [] # Multiple primers may go into the same destination, so need a new list to store these
        for mixture, destination_location in zip(self.dna_primer_mixtures, destination_locations):
            primers = mixture[1:]
            for primer_name in primers:
                primer_wells.append(primer_layout.get_wells_containing_liquid(primer_name)[0])
                primer_destination_locations.append(destination_location)

        Primer_Transfer_Source_Locations = _OTProto.get_locations(
                                                Labware = primer_labware,
                                                Wells = primer_wells
        )

        _OTProto.transfer_liquids(
                            Protocol = self._protocol,
                            Transfer_Volumes = Primer_Transfers,
                            Source_Locations = Primer_Transfer_Source_Locations,
                            Destination_Locations = primer_destination_locations,
                            new_tip = True,
                            mix_before = (3, "transfer_volume"),
                            mix_after = (3, "transfer_volume")
        )



from fileinput import filename
import json
import BiomationScripter as _BMS
import math
from opentrons import protocol_api
import warnings
from random import shuffle
import datetime
from typing import Literal, Union

BiomationScripter_Install_Dir = "/var/lib/jupyter/notebooks/Packages/BiomationScripterLib/BiomationScripter/OTProto/"
Pre_Loaded_Custom_Labware_Dir = BiomationScripter_Install_Dir + "Opentrons_Custom_Labware_Definitions/"

########################

class OTProto_Template:
    def __init__(self,
        Protocol: protocol_api,
        Name: str,
        Metadata: dict[ Literal['protocolName', 'author', 'author-email', 'user', 'user-email', 'source', 'apiLevel', 'robotName'], str ],
        Custom_Labware_Dir: str = "../Custom_Labware/",
        Starting_20uL_Tip: str = "A1",
        Starting_300uL_Tip: str = "A1",
        Starting_1000uL_Tip: str = "A1"
    ):
        #####################
        # Protocol Metadata #
        #####################
        self._protocol = Protocol
        self.name = Name
        self.metadata = Metadata
        self.custom_labware_dir = Custom_Labware_Dir

        ####################
        # Source materials #
        ####################
        ## Pipette Tips ##
        self.tip_types = {
            "p20": "opentrons_96_tiprack_20ul",
            "p300": "opentrons_96_tiprack_300ul",
            "p1000": "opentrons_96_tiprack_1000ul"
        }
        self.starting_tips = {
            "p20": Starting_20uL_Tip,
            "p300": Starting_300uL_Tip,
            "p1000": Starting_1000uL_Tip
        }

        self.tips_needed = {
            "p20": 0,
            "p300": 0,
            "p1000": 0
        }

        ###############
        # Robot Setup #
        ###############
        self._pipettes = {
            "left": "p20_single_gen2",
            "right": "p300_single_gen2"
        }
        self.__pipettes_loaded = False

    def pipettes_loaded(self):
        return(self.__pipettes_loaded)

    def custom_labware_directory(self, Directory):
        self.custom_labware_dir = Directory

    def pipette_config(self, Pipette_Type, Position):
        if self.__pipettes_loaded:
            raise _BMS.BMSTemplateError("Pipettes have already been loaded. Please modify pipette configuration before loading pipettes. Changes to configuration have NOT been saved.")

        if not Position.lower() in ["left", "right"]:
            raise _BMS.RobotConfigurationError("{} is an invalid pipette positon. Please specify either 'right' or 'left'".format(Position))

        self._pipettes[Position.lower()] = Pipette_Type


    def load_pipettes(self):
        for pipette_position in self._pipettes.keys():
            if pipette_position in self._protocol.loaded_instruments.keys():
                warnings.warn("Pipette already loaded in {} position so loading was skipped".format(pipette_position))
                continue
            pipette = self._pipettes[pipette_position]
            self._protocol.load_instrument(pipette, pipette_position)
        self.__pipettes_loaded = True

    def tip_type(self, Pipette, Tip_Type):
        try:
            self.tip_types[Pipette] = Tip_Type
        except KeyError:
            raise _BMS.RobotConfigurationError("{} is not a known pipette class. Please specify either p20, p300, or p1000".format(Pipette))

    def starting_tip_position(self, Pipette, Tip_Position):
        try:
            self.starting_tips[Pipette] = Tip_Position
        except:
            raise _BMS.RobotConfigurationError("{} is not a known pipette class. Please specify either p20, p300, or p1000".format(Pipette))

    def add_tip_boxes_to_pipettes(self):
        for pipette_type in self.tips_needed.keys():
            # Get the current pipette
            pipette = get_pipette(self._protocol, pipette_type)
            # Get the amount of tips needed for this pipette
            tips_needed = self.tips_needed[pipette_type]
            if tips_needed == 0:
                continue
            elif pipette == None and not self.tips_needed[pipette_type] == 0:
                raise _BMS.RobotConfigurationError("Tips have been specifed for {}, however this pipette type has not been loaded. Please ensure pipettes have been loaded and check your protocol.".format(pipette_type))
            else:
                tip_boxes_needed = tip_racks_needed(tips_needed, self.starting_tips[pipette_type])
                tip_type = self.tip_types[pipette_type]
                for tip_box in range(0, tip_boxes_needed):
                    tip_box_deck_slot = next_empty_slot(self._protocol)
                    tip_box = load_labware(self._protocol, tip_type, tip_box_deck_slot)

                    pipette.tip_racks.append(tip_box)

                pipette.starting_tip = pipette.tip_racks[0].well(self.starting_tips[pipette_type])

    def calculate_and_add_tips(self, Transfer_Volumes, New_Tip):
        Tips_20uL, Tips_300uL, Tips_1000uL = calculate_tips_needed(self._protocol, Transfer_Volumes, new_tip = New_Tip)
        self.tips_needed["p20"] += Tips_20uL
        self.tips_needed["p300"] += Tips_300uL
        self.tips_needed["p1000"] += Tips_1000uL

    def tip_racks_prompt(self):
        if not get_p20(self._protocol) == None:
            self._protocol.pause("This protocol uses {} 20 uL tip boxes".format(tip_racks_needed(self.tips_needed["p20"], self.starting_tips["p20"])))
        if not get_p300(self._protocol) == None:
            self._protocol.pause("This protocol uses {} 300 uL tip boxes".format(tip_racks_needed(self.tips_needed["p300"], self.starting_tips["p300"])))
        if not get_p1000(self._protocol) == None:
            self._protocol.pause("This protocol uses {} 1000 uL tip boxes".format(tip_racks_needed(self.tips_needed["p1000"], self.starting_tips["p1000"])))

    def run(self):
        #################
        # Load pipettes #
        #################
        self.load_pipettes()

    def run_as_module(self, Parent):
        #############################
        # Duplicate parent settings #
        #############################
        self.custom_labware_dir = Parent.custom_labware_dir
        self.tip_types = Parent.tip_types
        self._pipettes = Parent._pipettes
        self.__pipettes_loaded = True

        #############################
        # Get updated starting tips #
        #############################
        for pipette_type in self.starting_tips:
            pipette = get_pipette(self._protocol, pipette_type)
            if pipette:
                last_tip_rack = pipette.tip_racks[-1]
                next_tip_in_rack = last_tip_rack.next_tip().well_name
                self.starting_tips[pipette_type] = next_tip_in_rack

        ############################
        # Clear deck and tip racks #
        ############################
        for position in self._protocol.deck:
            if position == 12:
                # This is needed to stop the OT2 from deleting the waste bin
                # If the waste bin is deleted, there is a chance the tip will crash into
                # the side of it...
                continue
            del self._protocol.deck[position]

        # Clear the tip racks stored on each loaded pipette
        # This stops tip racks from being duplicated
        # Duplicated tip racks can cause the OT2 to try and use now-empty racks
        for pipette_type in self.tip_types:
            pipette = get_pipette(self._protocol, pipette_type)
            if pipette:
                pipette.tip_racks = []

        self.run()

########################

def get_locations(Labware, Wells, Direction = "Horizontal", Box = False):
    # Argument Direction is ignored is Wells is a list of wells
    ## It is only used if Wells is a well range (e.g. A1:H4)
    if not type(Wells) == list:
        if not ":" in Wells:
            Wells = [Wells]
        else:
            Wells = _BMS.well_range(
                        Wells = Wells,
                        Labware_Format = [len(Labware.rows()), len(Labware.columns())],
                        Direction = Direction,
                        Box = Box
            )

    Locations = []
    for well in Wells:
        Locations.append(Labware.wells_by_name()[well])

    return(Locations)

def get_pipette(Protocol, Pipette):
    if Pipette == "p20":
        return(get_p20(Protocol))
    elif Pipette == "p300":
        return(get_p300(Protocol))
    elif Pipette == "p1000":
        return(get_p1000(Protocol))
    else:
        raise _BMS.RobotConfigurationError("{} is not a known pipette class. Please specify either p20, p300, or p1000".format(Pipette))

def select_pipette_by_volume(Protocol, Volume):

    # Check which pipettes are available - any unavailable will return None
    p20 = get_p20(Protocol)
    p300 = get_p300(Protocol)
    p1000 = get_p1000(Protocol)

    if Volume < 1:
        raise _BMS.RobotConfigurationError("A volume of {} was specified. Cannot transfer less than 1 uL.".format(Volume))
    elif Volume < 20 and p20:
        return(p20)
    elif Volume == 20 and p20:
        return(p20)
    elif Volume == 20 and not p20:
        return(p300)
    elif Volume > 20 and Volume <= 300 and p300:
        return(p300)
    elif Volume > 300 and p1000:
        return(p1000)
    elif Volume >= 100 and not p300 and p1000:
        return(p1000)
    elif Volume > 300 and not p1000 and p300:
        return(p300)
    elif p20 and not p300 and not p1000:
        return(p20)
    else:
        raise _BMS.RobotConfigurationError("A suitable pipette is not loaded to transfer {} uL.\n Currently loaded pipettes:\n{}".format(Volume, Protocol._instruments))

def set_location_offset_top(Locations, Offset):
    # Check if a list of locations is given. If only one location is provided in non-list format, put it into a list
    if not type(Locations) == list:
        Locations = [Locations]

    Offset_Locations = []
    # For each location provided, apply the offset and add to Offset_Locations
    for location in Locations:
        Offset_Locations.append(location.top(Offset))

    return(Offset_Locations)

def set_location_offset_bottom(Locations, Offset):
    # Check if a list of locations is given. If only one location is provided in non-list format, put it into a list
    if not type(Locations) == list:
        Locations = [Locations]

    Offset_Locations = []
    # For each location provided, apply the offset and add to Offset_Locations
    for location in Locations:
        Offset_Locations.append(location.bottom(Offset))

    return(Offset_Locations)

def transfer_liquids(Protocol, Transfer_Volumes, Source_Locations, Destination_Locations, new_tip = True, mix_after = None, mix_before = None, mix_speed_multiplier = 1, aspirate_speed_multiplier = 1, dispense_speed_multiplier = 1, blowout_speed_multiplier = 1, touch_tip_source = False, touch_tip_destination = False, blow_out = False, blowout_location = "destination well", move_after_dispense = None):

    if not type(Transfer_Volumes) == list:
        Transfer_Volumes = [Transfer_Volumes]
    if not type(Source_Locations) == list:
        Source_Locations = [Source_Locations]
    if not type(Destination_Locations) == list:
        Destination_Locations = [Destination_Locations]

    # if not len(Transfer_Volumes) == len(Source_Locations) or not len(Transfer_Volumes) == len(Destination_Locations):
    #     raise ValueError("The number of transfer volumes, source locations, and destination locations are not the same.")

    # Get the smallest non-zero volume to transfer
    if Transfer_Volumes.count(0) == 0:
        min_transfer = min(Transfer_Volumes)
    else:
        temp_Transfer_Volumes = [tv for tv in Transfer_Volumes if not tv == 0]
        min_transfer = min(temp_Transfer_Volumes)

    # Get the largest volume to transfer
    max_transfer = max(Transfer_Volumes)

    # Check that the correct pipettes are available #
    p20 = get_p20(Protocol)
    p300 = get_p300(Protocol)
    p1000 = get_p1000(Protocol)

    # Modify the blowout flow rate (if needed)
    for pipette in [p20, p300, p1000]:
        if pipette:
            pipette.flow_rate.blow_out *= blowout_speed_multiplier

    if not p20 and not p300 and not p1000:
        raise ValueError("No pipettes have been loaded")

    for transfer_volume in Transfer_Volumes:
        if transfer_volume == 0:
            continue
        if transfer_volume < 20 and not p20:
            raise ValueError("p20 pipette is required, but has not been loaded")
        if transfer_volume == 20 and (not p20 and not p300):
            raise ValueError("p20 or p300 pipette is required, but neither have been loaded")
        if (transfer_volume <= 100 and transfer_volume > 20) and not p300:
            raise ValueError("p300 pipette is required, but has not been loaded")
        if transfer_volume <=300 and (not p300 and not p1000):
            raise ValueError("p300 or p1000 pipette is required, but neither have been loaded")

    ################################
    # Liquid Handling Instructions #
    ################################

    # If the same tip will be used for all transfers, select it here
    if not new_tip:
        ## May need to load a tip on each pipette to transfer all liquid volumes
        ### Start by selecting for the smallest transfer volume
        select_pipette_by_volume(Protocol, min_transfer).pick_up_tip()

        ### Then get tip for the largest volume
        # Check that the most suitable pipette doesn't already have a tip loaded for the smallest transfer
        if not select_pipette_by_volume(Protocol, max_transfer).has_tip:
            select_pipette_by_volume(Protocol, max_transfer).pick_up_tip()

    # Begin iterating through the transfer volumes list
    for transfer_volume, source, destination in zip(Transfer_Volumes, Source_Locations, Destination_Locations):
        # Ignore any transfers of 0 uL
        if transfer_volume == 0:
            continue

        # Select the best pipette for this transfer
        pipette = select_pipette_by_volume(Protocol, transfer_volume)

        # Load the pipette tip if required
        ## This will be needed if a new tip is being used for each transfer, and will be ignore if the same tip will always be used
        if new_tip:
            if pipette.has_tip:
                # Logic check - this should never occur
                ## If new_tip is `True`, the tip should have have been dropped after the previous transfer step
                raise _BMS.TransferError("Logical error for `OTProto.transfer_liquids()`:\nThe tip should have been dropped in the previous step but is still loaded. Please submit this protocol as an issue on the Github.")
            else:
                pipette.pick_up_tip()

        # The while loop is used to deal with situations where the transfer volume needs to be split into multiple transfers
        ## This can happen when the transfer volume is more than the largest pipette's max volume
        volume_transfered = 0 # used to track how much liquid has been transfered
        while not volume_transfered == transfer_volume:

            # Check if the transfer needs to be split, and set the volume to be transfered in this event
            if (transfer_volume - volume_transfered) > pipette.max_volume:
                current_transfer_volume = pipette.max_volume
            else:
                current_transfer_volume = transfer_volume - volume_transfered


            # Convert the mix_before and mix_after arguments into a useable fomrat
            mix_before_reps = None
            mix_before_volume = None

            if mix_before:
                mix_before_reps = mix_before[0]
                if mix_before[1] == "transfer_volume":
                    # Set the mix volume to the transfer volume
                    mix_before_volume = current_transfer_volume
                else:
                    # Set the mix volume to that provided by the user
                    mix_before_volume = mix_before[1]

            mix_after_reps = None
            mix_after_volume = None

            if mix_after:
                mix_after_reps = mix_after[0]
                if mix_after[1] == "transfer_volume":
                    # Set the mix volume to the transfer volume
                    mix_after_volume = current_transfer_volume
                else:
                    # Set the mix volume to that provided by the user
                    mix_after_volume = mix_after[1]


            # Move to the source location
            pipette.move_to(source.top())

            # Mix the source liquid (if required) #
            if mix_before:
                # Set the mixing speed
                pipette.flow_rate.aspirate *= mix_speed_multiplier
                pipette.flow_rate.dispense *= mix_speed_multiplier

                # Perform the mixing
                pipette.mix(
                    repetitions = mix_before_reps,
                    volume = mix_before_volume,
                    location = source
                )

                # Reset the flow rate speeds
                pipette.flow_rate.aspirate /= mix_speed_multiplier
                pipette.flow_rate.dispense /= mix_speed_multiplier

            # Aspirate the liquid #
            pipette.aspirate(
                volume = current_transfer_volume,
                location = source,
                rate = aspirate_speed_multiplier
            )

            # Knock the tip against the sides of the current well (if required)
            if touch_tip_source:
                pipette.touch_tip()

            # Move to the destination well
            pipette.move_to(destination.top())

            # Dispense the liquid
            pipette.dispense(
                volume = current_transfer_volume,
                location = destination,
                rate = dispense_speed_multiplier
            )

            # Mix the destination liquid (if required) #
            if mix_after:
                # Set the mixing speed
                pipette.flow_rate.aspirate *= mix_speed_multiplier
                pipette.flow_rate.dispense *= mix_speed_multiplier

                # Perform the mixing
                pipette.mix(
                    repetitions = mix_after_reps,
                    volume = mix_after_volume,
                    location = destination
                )

                # Reset the flow rate speeds
                pipette.flow_rate.aspirate /= mix_speed_multiplier
                pipette.flow_rate.dispense /= mix_speed_multiplier

            # Move the tip if required
            if move_after_dispense == "well_top":
                pipette.move_to(destination.top())
            elif move_after_dispense == "well_bottom":
                pipette.move_to(destination.bottom())
            elif move_after_dispense:
                # Check if the argument is not understood
                raise _BMS.TransferError("The `move_after_dispense` argument for `transfer_liquids` and `dispense_from_aliquots` MUST be either 'well_bottom' or 'well_top'. See the documentation for more information.")

            # If blowout will be somewhere other than the destination, then do touch_tip first
            if blowout_location == "source_well" or blowout_location == "trash":
                # knock the tip against the sides of the current well (if required)
                if touch_tip_destination:
                    pipette.touch_tip()

            # perform blowout if needed
            if blow_out:
                if blowout_location == "destination well":
                    b_location = destination
                elif blowout_location == "source_well":
                    b_location = source
                elif blowout_location == "trash":
                    b_location = Protocol.fixed_trash["A1"]
                else:
                    b_location = "" # this is to use the OT API error handling
                pipette.blow_out(b_location)

            # If blowout was at the destination location, then do blowout now

            if blowout_location == "destination well":
                # knock the tip against the sides of the current well (if required)
                if touch_tip_destination:
                    pipette.touch_tip()


            # If new_tip is True, drop the tip now
            if new_tip:
                pipette.drop_tip()

            # Update the volume left to transfer
            volume_transfered += current_transfer_volume

    # If the same tip was used for all transfers, then drop it now
    if not new_tip:
        select_pipette_by_volume(Protocol, min_transfer).drop_tip()
        # Check if the tip used for largest transfers has already been dropped (i.e. when the same pipette was used for everything)
        if select_pipette_by_volume(Protocol, max_transfer).has_tip:
            select_pipette_by_volume(Protocol, max_transfer).drop_tip()

    # Reset the blowout flow rate (if needed)
    for pipette in [p20, p300, p1000]:
        if pipette:
            pipette.flow_rate.blow_out /= blowout_speed_multiplier

def dispense_from_aliquots(Protocol, Transfer_Volumes, Aliquot_Source_Locations, Destinations, Min_Transfer = None, Calculate_Only = False, Dead_Volume_Proportion = 0.95, Aliquot_Volumes = None, new_tip = True, mix_after = None, mix_before = None, mix_speed_multiplier = 1, aspirate_speed_multiplier = 1, dispense_speed_multiplier = 1, blowout_speed_multiplier = 1, touch_tip_source = False, touch_tip_destination = False, blow_out = False, blowout_location = "destination well", move_after_dispense = None):

    Initial_Source_Locations = Aliquot_Source_Locations.copy()

    # If no min transfer specified, then use the min volume of the smallest loaded pipette
    if not Min_Transfer:
        Min_Transfer = get_lowest_volume_pipette(Protocol).min_volume

    # Create list to store order of Aliquot Source Location usage
    Aliquot_Source_Order = []

    # Format aliquot volumes so it is useable
    if Aliquot_Volumes:
        if type(Aliquot_Volumes) == int or type(Aliquot_Volumes) == float:
            Aliquot_Volumes = [Aliquot_Volumes] * len(Aliquot_Source_Locations)

    # Modify aliquot volumes to account for dead volumes
    if Aliquot_Volumes:
        Aliquot_Volumes = [vol*Dead_Volume_Proportion for vol in Aliquot_Volumes.copy()]

    Aliquot_Index = 0
    event_index = 0
    for transfer_volume, destination in zip(Transfer_Volumes, Destinations):
        # Check if there is enough volume in the current aliquot
        if Aliquot_Volumes:

            if Aliquot_Volumes[Aliquot_Index] >= transfer_volume:
                # Check that the vol being left behind isn't below the min transfer (unless this is the last transfer event)
                if not event_index == len(Transfer_Volumes) - 1:
                    if Aliquot_Volumes[Aliquot_Index] - transfer_volume >= Min_Transfer or Aliquot_Volumes[Aliquot_Index] - transfer_volume == 0: # second condition ensures that an aliquot can still be depleted
                        # print("Check", Aliquot_Volumes[Aliquot_Index] - transfer_volume, ">=", Min_Transfer)
                        # If all is fine, then continue
                        Aliquot_Volumes[Aliquot_Index] -= transfer_volume
                    else:
                        # Otherwise, leave enough behind that the aliquot can still be taken from (split the transfer)
                        new_transfer_vol = transfer_volume - Min_Transfer
                        # print("new transfer vol", new_transfer_vol)
                        ## Add the current aliquot source location to the source order
                        Aliquot_Source_Order.append(Aliquot_Source_Locations[Aliquot_Index])
                        # update the amount of liquid left in the aliquot
                        Aliquot_Volumes[Aliquot_Index] -= new_transfer_vol
                        # Iterate to the next aliquot
                        Aliquot_Index += 1
                        if Aliquot_Index == len(Aliquot_Source_Locations):
                            Aliquot_Index = 0
                        ## Insert the transfer volume and destination
                        Transfer_Volumes.insert(event_index, new_transfer_vol)
                        Destinations.insert(event_index, destination)
                        # Update the next transfer vol (i.e. the one that has just been split)
                        Transfer_Volumes[event_index + 1] -= new_transfer_vol
                        # Then move onto the next transfer action
                        event_index += 1
                        continue

            # If not, then split the transfer action
            else:
                # Transfer the rest of the current aliquot
                transfer_rest_vol = Aliquot_Volumes[Aliquot_Index]
                # print("To Transfer:", transfer_rest_vol, "<",  "Transfer Total", transfer_volume)

                # Check if the remaining volume to transfer is below the min transfer volume
                if transfer_volume - transfer_rest_vol < Min_Transfer:
                    # Change how much is being transfered in this first event so that the next transfer is above the minimum transfer amount
                    new_transfer_vol = transfer_volume - Min_Transfer
                    # print("new transfer vol", new_transfer_vol)
                    ## Add the current aliquot source location to the source order
                    Aliquot_Source_Order.append(Aliquot_Source_Locations[Aliquot_Index])
                    # update the amount of liquid left in the aliquot
                    Aliquot_Volumes[Aliquot_Index] -= new_transfer_vol
                    # Iterate to the next aliquot
                    Aliquot_Index += 1
                    if Aliquot_Index == len(Aliquot_Source_Locations):
                        Aliquot_Index = 0
                    ## Insert the transfer volume and destination
                    Transfer_Volumes.insert(event_index, new_transfer_vol)
                    Destinations.insert(event_index, destination)
                    # Update the next transfer vol (i.e. the one that has just been split)
                    Transfer_Volumes[event_index + 1] -= new_transfer_vol
                    # Then move onto the next transfer action
                    event_index += 1
                    continue

                # print("Left to transfer:", transfer_volume - transfer_rest_vol)
                ## Add the current aliquot source location to the source order
                Aliquot_Source_Order.append(Aliquot_Source_Locations[Aliquot_Index])
                ## And then remove it from the available list
                Aliquot_Volumes.remove(Aliquot_Volumes[Aliquot_Index])
                Aliquot_Source_Locations.remove(Aliquot_Source_Locations[Aliquot_Index])
                # If the aliquot removed was the last in the list, move back to the first aliquot
                if Aliquot_Index == len(Aliquot_Source_Locations):
                    Aliquot_Index = 0
                # If there aren't any aliquots left, raise an error
                if len(Aliquot_Source_Locations) == 0:
                    raise _BMS.OutOFSourceMaterial("Ran out of source material when aliquoting into {}.\nSource Locations:\n{}".format(destination, "\n".join(str(sl) for sl in Initial_Source_Locations)))

                ## Insert the transfer volume and destination
                Transfer_Volumes.insert(event_index, transfer_rest_vol)
                Destinations.insert(event_index, destination)
                # Update the next transfer vol (i.e. the one that has just been split)
                Transfer_Volumes[event_index + 1] -= transfer_rest_vol
                # print(Aliquot_Source_Locations[Aliquot_Index], destination, Transfer_Volumes[event_index + 1])
                # Then move onto the next transfer action
                event_index += 1
                continue

        Aliquot_Source_Order.append(Aliquot_Source_Locations[Aliquot_Index])
        Aliquot_Index += 1
        if Aliquot_Index == len(Aliquot_Source_Locations):
            Aliquot_Index = 0

        event_index += 1

    # Sanity check
    if not len(Transfer_Volumes) == len(Aliquot_Source_Order) and len(Transfer_Volumes) == len(Destinations):
        raise _BMS.TransferError("Internal error calculating dispense from aliquot transfer actions. Please report as an issue")

    if Calculate_Only:
        return(Transfer_Volumes, Aliquot_Source_Order, Destinations)
    else:
        transfer_liquids(
            Protocol = Protocol,
            Transfer_Volumes = Transfer_Volumes,
            Source_Locations = Aliquot_Source_Order,
            Destination_Locations = Destinations,
            new_tip = new_tip,
            mix_after = mix_after,
            mix_before = mix_before,
            mix_speed_multiplier = mix_speed_multiplier,
            aspirate_speed_multiplier = aspirate_speed_multiplier,
            dispense_speed_multiplier = dispense_speed_multiplier,
            blowout_speed_multiplier = blowout_speed_multiplier,
            touch_tip_source = touch_tip_source,
            touch_tip_destination = touch_tip_destination,
            blow_out = blow_out,
            blowout_location = blowout_location,
            move_after_dispense = move_after_dispense
        )

def next_empty_slot(protocol):
    # temporary workaround if Thermocycler is loaded
    if str(protocol.deck[7]) == "Thermocycler Module on 7":
        for slot in [1,2,3,4,5,6,9]:
            labware = protocol.deck[slot]
            if labware is None: # if no labware loaded into slot
                return(slot)
    else:
        for slot in protocol.deck:
            labware = protocol.deck[slot]
            if labware is None: # if no labware loaded into slot
                return(slot)
    raise IndexError('No Deck Slots Remaining')

def get_labware_format(labware_api_name, custom_labware_dir = None):
    # If try code block fails, labware may be custom, so treat it as such
    try:
        labware_definition = protocol_api.labware.get_labware_definition(labware_api_name)
        n_cols = len(labware_definition["ordering"])
        n_rows = len(labware_definition["ordering"][0])
        return(n_rows, n_cols)

    except FileNotFoundError:
        try:
            definition_file_location = "{}/{}.json".format(custom_labware_dir, labware_api_name)
            with open(definition_file_location) as labware_definition:
                data = json.load(labware_definition)
                n_cols = len(data["ordering"])
                n_rows = len(data["ordering"][0])
                return(n_rows, n_cols)
        except FileNotFoundError:
            definition_file_location = "{}/{}.json".format(Pre_Loaded_Custom_Labware_Dir, labware_api_name)
            with open(definition_file_location) as labware_definition:
                data = json.load(labware_definition)
                n_cols = len(data["ordering"])
                n_rows = len(data["ordering"][0])
                return(n_rows, n_cols)

def get_labware_well_capacity(labware_api_name, custom_labware_dir = None):
    # If try code block fails, labware may be custom, so treat it as such
    try:
        labware_definition = protocol_api.labware.get_labware_definition(labware_api_name)
        # Check the capacity of all wells in the labware and add to a set
        capacities = set()
        for well in labware_definition["wells"]:
            capacities.add(labware_definition["wells"][well]["totalLiquidVolume"])
        # Check if the capacities set has more than one volume
        ## If it does, raise an error becuase I can't think of an easy way to deal with that atm
        if not len(capacities) == 1:
            raise _BMS.LabwareError("Labware {} has slots/wells with different volume capacities; BMS cannot currently deal with this.")
        capacity = capacities.pop()
        return(capacity)

    except FileNotFoundError:
        try:
            definition_file_location = "{}/{}.json".format(custom_labware_dir, labware_api_name)
            with open(definition_file_location) as labware_definition:
                data = json.load(labware_definition)
                # Check the capacity of all wells in the labware and add to a set
                capacities = set()
                for well in data["wells"]:
                    capacities.add(data["wells"][well]["totalLiquidVolume"])
                # Check if the capacities set has more than one volume
                ## If it does, raise an error becuase I can't think of an easy way to deal with that atm
                if not len(capacities) == 1:
                    raise _BMS.LabwareError("Labware {} has slots/wells with different volume capacities; BMS cannot currently deal with this.")
                capacity = capacities.pop()
                return(capacity)
        except FileNotFoundError:
            definition_file_location = "{}/{}.json".format(Pre_Loaded_Custom_Labware_Dir, labware_api_name)
            with open(definition_file_location) as labware_definition:
                data = json.load(labware_definition)
                # Check the capacity of all wells in the labware and add to a set
                capacities = set()
                for well in data["wells"]:
                    capacities.add(data["wells"][well]["totalLiquidVolume"])
                # Check if the capacities set has more than one volume
                ## If it does, raise an error becuase I can't think of an easy way to deal with that atm
                if not len(capacities) == 1:
                    raise _BMS.LabwareError("Labware {} has slots/wells with different volume capacities; BMS cannot currently deal with this.")
                capacity = capacities.pop()
                return(capacity)

def load_labware(parent, labware_api_name, deck_position = None, custom_labware_dir = None, label = None):
    labware = None

    # Try and load not as custom
    try:
        # Check if `parent` is the deck or a hardware module, and treat it acordingly
        if parent.__class__ == protocol_api.protocol_context.ProtocolContext:
            # If no deck position, get the next empty slot
            if not deck_position:
                deck_position = next_empty_slot(parent)
            labware = parent.load_labware(labware_api_name, deck_position, label)
        else:
            labware = parent.load_labware(labware_api_name, label)
    except:
        if not custom_labware_dir:
            raise ValueError("{} appears to be custom labware; use `custom_labware_dir` to provide the directory location for this labware file".format(labware_api_name))
        labware = load_custom_labware(parent, custom_labware_dir + "/" + labware_api_name + ".json", deck_position, label)

    if labware == None:
        raise ValueError("Labware not loaded for unknown reasons.")

    return(labware)

def calculate_and_load_labware(protocol, labware_api_name, wells_required, wells_available = "all", modules=None, custom_labware_dir = None, label = None):
    # Determine amount of labware required #
    labware = []
    ## Load first labware to get format - assume always at least one required
    # load directly onto modules if available
    if modules is None:
        labware_slot = next_empty_slot(protocol)
        loaded_labware = load_labware(protocol, labware_api_name, labware_slot, custom_labware_dir = custom_labware_dir, label = label)
    else:
        if type(modules) is not list:
            modules = [modules]
        loaded_labware = load_labware(modules[0], labware_api_name, custom_labware_dir = custom_labware_dir, label = label)
    labware.append(loaded_labware)
    ## Determine space in labware
    if wells_available == "all":
        wells_in_labware = len(labware[0].wells())
    else:
        wells_in_labware = wells_available
    ## Determine total amount of dilution labware required
    n_labware = math.ceil(wells_required/wells_in_labware)
    ## Load more labware if required
    for lw in range(0, n_labware - 1):
        # load directly onto modules if available
        if modules is not None and len(modules) > lw+1:
            loaded_labware = load_labware(modules[lw+1], labware_api_name, custom_labware_dir = custom_labware_dir, label = label)
            labware.append(loaded_labware)
        else:
            labware_slot = next_empty_slot(protocol)
            loaded_labware = load_labware(protocol, labware_api_name, labware_slot, custom_labware_dir = custom_labware_dir, label = label)
            labware.append(loaded_labware)

    well_locations = []
    labware_index = 0
    location_in_labware_index = 0
    for well in range(0,wells_required):
        current_labware_wells = labware[labware_index].wells()
        well_locations.append(current_labware_wells[location_in_labware_index])
        location_in_labware_index += 1
        if location_in_labware_index == len(current_labware_wells):
            labware_index += 1
            location_in_labware_index = 0

    # Return the list of loaded labware
    return(labware, well_locations)

def load_labware_from_layout(Protocol, Labware_Layout, deck_position = None, custom_labware_dir = None):
    labware_type = Labware_Layout.type
    labware_name = Labware_Layout.name
    labware = load_labware(Protocol, labware_type, deck_position = deck_position, custom_labware_dir = custom_labware_dir, label = labware_name)

    return(labware)

def calculate_tips_needed(protocol, transfers, template = None, new_tip = True):
    if not type(transfers) == list:
        transfers = [transfers]

    tips_needed = {
        "20uL": 0,
        "300uL": 0,
        "1000uL": 0
    }

    for transfer in transfers:
        # If transfer volume is 0 uL, skip it
        if transfer == 0:
            continue
        # Get the most appropriate pipette from those which are loaded
        pipette = select_pipette_by_volume(protocol, transfer)
        # Check how many trasfers will be needed to transfer the entire volume (e.g. if p300 is chosen for 400 uL, two transfers needed)
        ## This can be ignored if new_tip == False (i.e. one tip will be used no matter hwhat)
        if new_tip:
            transfers_needed = math.ceil(transfer/pipette.max_volume)
            # For each transfer that is needed, add to the appropriate tips_needed counter
            for n in range(0, transfers_needed):
                tips_needed["{}uL".format(pipette.max_volume)] += 1
        else:
            # Check if a new tip needs to be added
            ## A new tip won't be needed if an appropriate sized tip has already been added and new_tip is False
            if tips_needed["{}uL".format(pipette.max_volume)] > 0:
                continue
            else:
                tips_needed["{}uL".format(pipette.max_volume)] += 1


    # If a template is specified, add the tips needed to the tips_needed attributes
    if template:
        template.tips_needed["p20"] += tips_needed["20uL"]
        template.tips_needed["p300"] += tips_needed["300uL"]
        template.tips_needed["p1000"] += tips_needed["1000uL"]

    # Return a list of all tips needed
    return(tips_needed["20uL"], tips_needed["300uL"], tips_needed["1000uL"])

def tip_racks_needed(tips_needed, starting_tip_position = "A1"):
    if tips_needed == 0:
        return(0)
    tips_in_first_rack = len(_BMS.well_range("{}:H12".format(starting_tip_position), [8,12], "Vertical"))
    if tips_needed > tips_in_first_rack:
        extra_racks_required = math.ceil((tips_needed - tips_in_first_rack)/96)
    else:
        extra_racks_required = 0
    racks_required = 1 + extra_racks_required
    return(racks_required)

def load_pipettes_and_tips(Protocol, Pipette_Type, Pipette_Position, Tip_Type, Number_Tips_Required = None, Starting_Tip = "A1"):
    ## When Number_Tips_Required is False, just one tip box is created and asigned to the pipette
    tip_racks = []

    if Number_Tips_Required:
        n_tip_racks = tip_racks_needed(Number_Tips_Required, Starting_Tip)
    else:
        n_tip_racks = 1

    for tip_rack in range(0, n_tip_racks):
        tip_rack_deck_slot = next_empty_slot(Protocol)
        tip_rack = load_labware(Protocol, Tip_Type, tip_rack_deck_slot)
        tip_racks.append(tip_rack)

    # Check if pipette is already loaded
    # if something is loaded but it isn't the same type as the one being loaded raise an error
    # if something is loaded and it is the same type as the one beign loaded, pass
    # if something isn't loaded, then load it
    if Pipette_Position in Protocol.loaded_instruments.keys():
        if Protocol.loaded_instruments[Pipette_Position].name == Pipette_Type:
            pipette = Protocol.loaded_instruments[Pipette_Position]
            for tip_rack in tip_racks:
                pipette.tip_racks.append(tip_rack)
        else:
            raise _BMS.RobotConfigurationError("A pipette is already loaded, check the protocol for errors.")
    else:
        pipette = Protocol.load_instrument(Pipette_Type, Pipette_Position, tip_racks)

    pipette.starting_tip = tip_racks[0].well(Starting_Tip)

    return(pipette, tip_racks)

def add_tip_boxes_to_pipettes(Protocol, Pipette_Type, Tip_Type, Tips_Needed, Starting_Tip = "A1"):
    if Tips_Needed == 0:
        return(None)
    pipette = get_pipette(Protocol, Pipette_Type)
    if pipette == None:
        raise _BMS.RobotConfigurationError("The specified pipette type, {}, has not been loaded.".format(Pipette_Type))
    tip_boxes_needed = tip_racks_needed(Tips_Needed, Starting_Tip)
    for tip_box in range(0, tip_boxes_needed):
        tip_box_deck_slot = next_empty_slot(Protocol)
        tip_box = load_labware(Protocol, Tip_Type, tip_box_deck_slot)

        pipette.tip_racks.append(tip_box)
    pipette.starting_tip = pipette.tip_racks[0].well(Starting_Tip)



# Functions less likely to be useful to users

def get_p20(protocol):
    pipettes = protocol.loaded_instruments
    for position in pipettes.keys():
        min_volume = pipettes[position].min_volume
        max_volume = pipettes[position].max_volume
        channels = pipettes[position].channels
        if (min_volume == 1) and (max_volume == 20) and (channels == 1):
            return(pipettes[position])
    return(None)

def get_p300(protocol):
    pipettes = protocol.loaded_instruments
    for position in pipettes.keys():
        min_volume = pipettes[position].min_volume
        max_volume = pipettes[position].max_volume
        channels = pipettes[position].channels
        if (min_volume == 20) and (max_volume == 300) and (channels == 1):
            return(pipettes[position])
    return(None)

def get_p1000(protocol):
    pipettes = protocol.loaded_instruments
    for position in pipettes.keys():
        min_volume = pipettes[position].min_volume
        max_volume = pipettes[position].max_volume
        channels = pipettes[position].channels
        if (min_volume == 100) and (max_volume == 1000) and (channels == 1):
            return(pipettes[position])
    return(None)

def get_lowest_volume_pipette(protocol):
    if get_pipette(protocol, "p20"):
        return(get_pipette(protocol, "p20"))
    elif get_pipette(protocol, "p300"):
        return(get_pipette(protocol, "p300"))
    elif get_pipette(protocol, "p1000"):
        return(get_pipette(protocol, "p1000"))

    return(None)

def load_custom_labware(parent, file, deck_position = None, label = None):
    # Open the labware json file
    with open(file) as labware_file:
        labware_file = json.load(labware_file)

    # Check if `parent` is the deck or a hardware module, and treat it acordingly
    if parent.__class__ == protocol_api.protocol_context.ProtocolContext:
        # If no deck position, get the next empty slot
        if not deck_position:
            deck_position = next_empty_slot(parent)
        labware = parent.load_labware_from_definition(labware_file, deck_position, label)
    else:
        labware = parent.load_labware_from_definition(labware_file, label)

    return(labware)

def shuffle_locations(protocol, Locations, outdir=None, outfile="Locations"):
    old_locations = Locations.copy()
    # shuffle the locations
    shuffle(Locations)
    new_locations = Locations
    # add the locations to the robot command line to be logged and seen by the user
    protocol.comment(f"Locations given were {old_locations}")
    protocol.comment(f"Shuffled locations are {new_locations}")
    # export to file (if outdir provided)
    # recommend user to provide outdir="/data/user_storage/<USER_DIR>"
    # this function will overwrite any existing file with the same path
    if outdir is not None:
        fw = open(outdir+outfile+".csv", "w")
        dt = datetime.datetime.now()
        fw.write(dt.strftime("%c"))
        fw.write("\nOld Location,New Location,\n")
        for o,n in zip(old_locations,new_locations):
            fw.write(f"{o},{n},\n")
        fw.close()
    return new_locations



import BiomationScripter as _BMS
import BiomationScripter.OTProto as _OTProto
from opentrons import simulate as OT2
import math
import warnings
from typing import List, NewType, Tuple, Union, Literal
from collections import namedtuple


class Template(_OTProto.OTProto_Template):
    def __init__(self,
        DNA_Source_Layouts: List[_BMS.Labware_Layout],
        Competent_Cells_Source_Type: str,
        Transformation_Destination_Type: str,
        Media_Source_Type: str,
        DNA_Volume_Per_Transformation: float,
        Competent_Cell_Volume_Per_Transformation: float,
        Transformation_Final_Volume: float,
        Heat_Shock_Time: int,
        Heat_Shock_Temp: int,
        Media_Aliquot_Volume: float,
        Competent_Cells_Aliquot_Volume: float,
        Wait_Before_Shock: int,
        Replicates: int,
        Heat_Shock_Modules: List[str] = ["temperature module gen2"],
        Cooled_Cells_Modules: List[str] = [],
        Shuffle: Union[Tuple[str, str] , None] = None,
        Patience: int = 1200,
        Cells_Mix_Before: Union[Tuple[int, Union[float , Literal["transfer_volume"]]] , None] = (5,"transfer_volume"),
        Cells_Mix_After: Union[Tuple[int, Union[float , Literal["transfer_volume"]]] , None] = None,
        DNA_Mix_Before: Union[Tuple[int, Union[float , Literal["transfer_volume"]]] , None] = None,
        DNA_Mix_After: Union[Tuple[int, Union[float , Literal["transfer_volume"]]] , None] = (10,"transfer_volume"),
        **kwargs
    ):

        ########################################
        # User defined aspects of the protocol #
        ########################################
        self.dna_per_transformation = DNA_Volume_Per_Transformation
        self.cells_per_transformation = Competent_Cell_Volume_Per_Transformation
        self.final_volume = Transformation_Final_Volume
        self.heat_shock_time = Heat_Shock_Time # seconds
        self.heat_shock_temp = Heat_Shock_Temp # celcius
        self.wait_before_shock = Wait_Before_Shock # seconds
        self.replicates = Replicates
        self.shuffle = Shuffle
        self.patience = Patience
        self.cells_mix_before = Cells_Mix_Before
        self.cells_mix_after = Cells_Mix_After
        self.dna_mix_before = DNA_Mix_Before
        self.dna_mix_after = DNA_Mix_After

        self.heat_shock_modules = Heat_Shock_Modules
        self.cooled_cells_modules = Cooled_Cells_Modules

        ####################
        # Source materials #
        ####################
        self.dna_source_layouts = DNA_Source_Layouts

        self.comp_cells_source_type = Competent_Cells_Source_Type
        self.comp_cells_aliquot_volume = Competent_Cells_Aliquot_Volume

        self.media_source_type = Media_Source_Type
        self.media_aliquot_volume = Media_Aliquot_Volume

        #######################
        # Destination Labware #
        #######################
        self.destination_type = Transformation_Destination_Type

        ##################################################
        # Protocol Metadata and Instrument Configuration #
        ##################################################
        self._temperature_module = "temperature module gen2"
        self._tc_module = "Thermocycler Module"

        self._module_positions = [1, 3, 4, 6, 7, 9, 10]

        super().__init__(**kwargs)

    def run(self):
        #################
        # Load pipettes #
        #################
        self.load_pipettes()

        ################
        # Load Modules #
        ################

        # Load the destination modules
        destination_modules = []

        # Check if the thermocycler will be used - load it first if it will be
        if self._tc_module in self.heat_shock_modules:
            destination_modules.append(self._protocol.load_module(self._tc_module))
            self._module_positions = [1, 3, 4, 6, 10]

        # Then load any remaining destination (heat shock) modules
        for hs_mod in self.heat_shock_modules:
            # Check that it isn't the thermocyler (if applicable)
            if not self._tc_module == hs_mod:
                # Get a suitable deck position
                deck_slot = _OTProto.next_empty_slot(self._protocol)
                while not deck_slot in self._module_positions and not deck_slot in self._module_positions:
                    deck_slot += 1
                    if deck_slot >= 12:
                        raise _BMS.LabwareError("Not enough deck positions on the robot.")
                # Load the module
                destination_modules.append(self._protocol.load_module(hs_mod, deck_slot))

        # Load the source modules
        source_modules = []
        for cc_mod in self.cooled_cells_modules:
            # Get a suitable deck position
            deck_slot = _OTProto.next_empty_slot(self._protocol)
            labware = self._protocol.deck[deck_slot]
            while not deck_slot in self._module_positions:
                deck_slot += 1
                if deck_slot >= 12:
                    raise _BMS.LabwareError("Not enough deck positions on the robot.")
                labware = self._protocol.deck[deck_slot]
                # check that the deck slot is empty
                while labware is not None:
                    deck_slot += 1
                    if deck_slot >= 12:
                        raise _BMS.LabwareError("Not enough deck positions on the robot.")
                    labware = self._protocol.deck[deck_slot]
            # Load the module
            source_modules.append(self._protocol.load_module(cc_mod, deck_slot))

        if len(source_modules) == 0:
            source_modules = None


        ################################
        # Create transfer volume lists #
        ################################
        Num_Transformations = sum([len(dl.get_occupied_wells()) for dl in self.dna_source_layouts]) * self.replicates

        Cell_Transfer_Volumes = [self.cells_per_transformation] * Num_Transformations
        DNA_Transfer_Volumes = [self.dna_per_transformation] * Num_Transformations
        Media_Transfer_Volumes = [self.final_volume - self.dna_per_transformation - self.cells_per_transformation] * Num_Transformations

        #########################################################################
        # Calculate the number of tips and tip racks required for this protocol #
        #########################################################################

        self.calculate_and_add_tips(Cell_Transfer_Volumes, New_Tip = False)
        self.calculate_and_add_tips(DNA_Transfer_Volumes, New_Tip = True)
        self.calculate_and_add_tips(Media_Transfer_Volumes, New_Tip = True)
        self.add_tip_boxes_to_pipettes() # Starting tip(s) are defined here as well

        #######################
        # Load Source Labware #
        #######################
        # Competent Cells
        Cell_Aliquots_Required = math.ceil(sum(Cell_Transfer_Volumes)/self.comp_cells_aliquot_volume)
        Cell_Source_Labware, Cell_Source_Locations = _OTProto.calculate_and_load_labware(self._protocol, self.comp_cells_source_type, Cell_Aliquots_Required, modules=source_modules, custom_labware_dir = self.custom_labware_dir)

        # DNA
        DNA_Source_Labware = [
            _OTProto.load_labware_from_layout(
                Protocol = self._protocol,
                Labware_Layout = dl,
                custom_labware_dir = self.custom_labware_dir
            )
            for dl in self.dna_source_layouts
        ]

        layout_labware_mapping = {}

        for layout, labware in zip(self.dna_source_layouts, DNA_Source_Labware):
            layout_labware_mapping[labware] = layout

        DNA_Source_Locations = []

        for layout, labware in zip(self.dna_source_layouts, DNA_Source_Labware):
            DNA_Source_Locations += _OTProto.get_locations(
                Labware = labware,
                Wells = layout.get_occupied_wells()
            )

        # Shuffle DNA locations
        if self.shuffle is not None:
            outD = self.shuffle[0]
            outF = self.shuffle[1]
            DNA_Source_Locations = _OTProto.shuffle_locations(self._protocol, DNA_Source_Locations, outdir=outD, outfile=outF)

        # Media
        # if media arguments are set to None, the media transfer steps are skipped
        if self.media_aliquot_volume is not None:
            Media_Aliquots_Required = math.ceil(sum(Media_Transfer_Volumes)/self.media_aliquot_volume)
            Media_Source_Labware, Media_Source_Locations = _OTProto.calculate_and_load_labware(self._protocol, self.media_source_type, Media_Aliquots_Required, custom_labware_dir = self.custom_labware_dir)

        ############################
        # Load Destination Labware #
        ############################
        Destination_Labware, Destination_Locations = _OTProto.calculate_and_load_labware(self._protocol, self.destination_type, Num_Transformations, modules=destination_modules, custom_labware_dir = self.custom_labware_dir)

        print("Transformation Mapping")
        for dna, destination in zip([f"{layout_labware_mapping[location.parent].get_liquids_in_well(location.well_name)[0]} ({location})" for location in DNA_Source_Locations], Destination_Locations):
            print(f"{dna} -> {destination}")

        ######################
        # User Setup Prompts #
        ######################
        self.tip_racks_prompt()

        for dna_name, location in  zip([layout.get_liquids_in_well(well)[0] for layout in self.dna_source_layouts for well in layout.get_occupied_wells()], DNA_Source_Locations):
            self._protocol.pause("Place DNA Sample {} at {}".format(dna_name, location))

        if self.media_aliquot_volume is not None:
            self._protocol.pause("This protocol uses {} aliquots of {} uL media, located at {}".format(Media_Aliquots_Required, self.media_aliquot_volume, Media_Source_Locations))
        else:
            self._protocol.pause("Media Aliquot Volume was set to None, media transfer steps shall be skipped")
        self._protocol.pause("This protocol uses {} aliquots of {} uL competent cells, located at {}".format(Cell_Aliquots_Required, self.comp_cells_aliquot_volume, Cell_Source_Locations))

        ##########################
        # Liquid handling begins #
        ##########################

        # Set temperature to 4C
        for hs_mod in destination_modules:
            if hs_mod._module.model() == "thermocyclerModuleV1":
                hs_mod.set_block_temperature(4)
                hs_mod.open_lid()
            else:
                hs_mod.start_set_temperature(4)

        if source_modules:
            for cc_mod in source_modules:
                cc_mod.start_set_temperature(4)

        # Add comp cells
        _OTProto.dispense_from_aliquots(
            self._protocol,
            Cell_Transfer_Volumes,
            Cell_Source_Locations,
            Destination_Locations,
            Min_Transfer = None,
            Calculate_Only = False,
            Dead_Volume_Proportion = 1,
            Aliquot_Volumes = self.comp_cells_aliquot_volume,
            new_tip = False,
            mix_after = self.cells_mix_after,
            mix_before = self.cells_mix_before,
            mix_speed_multiplier = 1.5,
            aspirate_speed_multiplier = 1,
            dispense_speed_multiplier = 1,
            blowout_speed_multiplier = 1,
            touch_tip_source = False,
            touch_tip_destination = True,
            blow_out = True,
            blowout_location = "destination well",
            move_after_dispense = "well_bottom"
        )

        # Add DNA
        _OTProto.transfer_liquids(
            self._protocol,
            DNA_Transfer_Volumes,
            DNA_Source_Locations,
            Destination_Locations,
            new_tip = True,
            mix_after = self.dna_mix_after,
            mix_before = self.dna_mix_before,
            mix_speed_multiplier = 2,
            aspirate_speed_multiplier = 1,
            dispense_speed_multiplier = 1,
            blowout_speed_multiplier = 1,
            touch_tip_source = False,
            touch_tip_destination = True,
            blow_out = True,
            blowout_location = "destination well",
            move_after_dispense = "well_bottom"
        )

        # Heat shock

        # Wait for a bit
        self._protocol.delay(seconds = self.wait_before_shock)

        for hs_mod in destination_modules:
            if hs_mod._module.model() == "thermocyclerModuleV1":
                # close thermocycler lid
                hs_mod.close_lid()
                # heat shock as user specified
                hs_mod.set_block_temperature(self.heat_shock_temp, hold_time_seconds=self.heat_shock_time)
                # hold at 4 for 2 minutes
                hs_mod.set_block_temperature(4, 120)
                # open the lid
                hs_mod.open_lid()
            else:
                hs_mod.set_temperature(self.heat_shock_temp)
                # Wait for a bit
                self._protocol.delay(seconds = self.heat_shock_time)
                # Cool back to 4
                hs_mod.start_set_temperature(4)

        # Ensure all modules get held at 4c for 2 mins
        for hs_mod in destination_modules:
            if hs_mod._module.model() != "thermocyclerModuleV1":
                timer = 0
                while hs_mod._module.status == "cooling" and timer < self.patience:
                    self._protocol.delay(seconds = 5)
                    timer += 5
        self._protocol.delay(seconds = 120)


        # # Set the temp to heat shock
        # if tc_module is not None:
        #     # close thermocycler lid
        #     tc_module.close_lid()
        #     # heat shock as user specified
        #     tc_module.set_block_temperature(self.heat_shock_temp, hold_time_seconds=self.heat_shock_time)
        #     # hold at 4 for 2 minutes
        #     tc_module.set_block_temperature(4, hold_time_seconds=120)
        #     # open the lid
        #     tc_module.open_lid()
        # else:
        #     destination_module.set_temperature(self.heat_shock_temp)
        #     # Wait for a bit
        #     self._protocol.delay(seconds = self.heat_shock_time)
        #     # Cool back to 4 - protocol won't continue until this is back at 4...
        #     destination_module.set_temperature(4)

        # Add media
        if self.media_aliquot_volume is not None:
            # Prompt user to open LB tubes
            self._protocol.pause("Open up LB tubes")

            _OTProto.dispense_from_aliquots(
                self._protocol,
                Media_Transfer_Volumes,
                Media_Source_Locations,
                Destination_Locations,
                Min_Transfer = None,
                Calculate_Only = False,
                Dead_Volume_Proportion = 1,
                Aliquot_Volumes = self.media_aliquot_volume,
                new_tip = True,
                mix_after = None,
                mix_before = None,
                mix_speed_multiplier = 1,
                aspirate_speed_multiplier = 1,
                dispense_speed_multiplier = 1,
                blowout_speed_multiplier = 1,
                touch_tip_source = False,
                touch_tip_destination = True,
                blow_out = True,
                blowout_location = "destination well",
                move_after_dispense = None
            )



import BiomationScripter as _BMS
import BiomationScripter.OTProto as _OTProto
from opentrons import simulate as OT2
import math
import warnings
from typing import List, NewType, Tuple, Union, Literal
from collections import namedtuple


class Template(_OTProto.OTProto_Template):
    def __init__(self,
        Final_fmol: float,
        DNA: List[str],
        DNA_Concentration: List[float],
        DNA_Length: List[int],
        DNA_Source_Type: str,
        Keep_In_Current_Wells: bool,
        Water_Source_Labware_Type: str,
        Water_Aliquot_Volume: float,
        DNA_Source_Wells: list[str] = None,
        Final_Volume: float = None,
        Current_Volume: List[float] = None,
        Destination_Labware_Type: str = None,
        Destination_Labware_Wells: List[str] = None,
        **kwargs
    ):

        ########################################
        # User defined aspects of the protocol #
        ########################################
        self.final_fmol = Final_fmol
        self._keep_in_current_wells = Keep_In_Current_Wells
        self.final_volume = Final_Volume

        ####################
        # Source materials #
        ####################
        ## DNA to be diluted ##
        self.dna = DNA
        self.dna_source_wells = DNA_Source_Wells
        self.dna_source_type = DNA_Source_Type
        self.dna_starting_volume = Current_Volume
        self.dna_starting_concentration = DNA_Concentration
        self.dna_length = DNA_Length
        ## Water ##
        self.water_source_labware_type = Water_Source_Labware_Type
        self.water_aliquot_volume = Water_Aliquot_Volume

        #######################
        # Destination Labware #
        #######################
        self._destination_labware_type = Destination_Labware_Type
        self._destination_labware_wells = Destination_Labware_Wells

        ##################################################
        # Protocol Metadata and Instrument Configuration #
        ##################################################
        super().__init__(**kwargs)

    def run(self):
        #################
        # Load pipettes #
        #################
        self.load_pipettes()

        #################################################
        # Calculate dilution factor for each DNA sample #
        #################################################
        dna_current_fmol = []
        dna_dilution_factor = []
        for dna_name, dna_starting_concentration, dna_length in zip(self.dna, self.dna_starting_concentration, self.dna_length):
            mass_g = dna_starting_concentration * 1e-9
            length = dna_length
            fmol = (mass_g/((length * 617.96) + 36.04)) * 1e15
            dna_current_fmol.append(fmol)
            dilution_factor = self.final_fmol/fmol
            dna_dilution_factor.append(dilution_factor)
            if dilution_factor > 1:
                raise _BMS.NegativeVolumeError("DNA sample {} is too dilute and is already below {} fmol/uL".format(dna_name, self.final_fmol))

        ####################################################################
        # Determine if DNA will be diluted in current wells or transferred #
        ####################################################################
        if self._keep_in_current_wells == True:
            # Calculate amount of water to add to each DNA sample #
            water_to_add = []
            for starting_volume, dilution_factor in zip(self.dna_starting_volume, dna_dilution_factor):
                water_to_add.append((starting_volume / dilution_factor) - starting_volume)

            # Calculate number of tips needed:
            self.tips_needed["p20"], self.tips_needed["p300"], self.tips_needed["p1000"] = _OTProto.calculate_tips_needed(self._protocol, water_to_add, new_tip = True)
            # Add required number of tip boxes to the loaded pipettes
            self.add_tip_boxes_to_pipettes()

            # Load source labware #
            DNA_labware = _OTProto.load_labware(self._protocol, self.dna_source_type, custom_labware_dir = self.custom_labware_dir, label = "DNA Source Labware")
            ## Calculate number of water aliquots required
            total_water_required = sum(water_to_add)
            aliquots_required = math.ceil(total_water_required/self.water_aliquot_volume)
            ## Load required water source labware
            water_source_labware, water_source_locations = _OTProto.calculate_and_load_labware(self._protocol, self.water_source_labware_type, aliquots_required, custom_labware_dir = self.custom_labware_dir)

            ## DNA source labware is also the destination labware in this case
            ## But get DNA locations
            DNA_Locations = []
            # If DNA source wells were not specified, then choose wells to use
            if self.dna_source_wells == None:
                if len(DNA_labware.wells_by_name()) < len(self.dna):
                    raise _BMS.BMSTemplateError("Cannot currently use two source labware.")
                self.dna_source_wells = []
                for dna, source_well in zip(self.dna, DNA_labware.wells_by_name()):
                    self.dna_source_wells.append(source_well)

            for dna_well in self.dna_source_wells:
                DNA_Locations.append(DNA_labware.wells_by_name()[dna_well])

            # Liquid handling begins #
            ## Dispense water into DNA source wells and mix
            _OTProto.dispense_from_aliquots(self._protocol, water_to_add, water_source_locations, DNA_Locations, new_tip = True, mix_after = (5,"transfer_volume"))

        else:
            # Calculate amount of water to add to each DNA sample #
            water_to_add = []
            dna_to_add = []
            for dilution_factor in dna_dilution_factor:
                final_volume = self.final_volume
                dna_volume = final_volume * dilution_factor
                water_volume = final_volume - dna_volume
                water_to_add.append(water_volume)
                dna_to_add.append(dna_volume)

            # Calculate number of tips needed:
            water_tips_needed_20uL, water_tips_needed_300uL, water_tips_needed_1000uL = _OTProto.calculate_tips_needed(self._protocol, water_to_add, new_tip = False)
            dna_tips_needed_20uL, dna_tips_needed_300uL, dna_tips_needed_1000uL = _OTProto.calculate_tips_needed(self._protocol, dna_to_add, new_tip = True)

            self.tips_needed["p20"] = water_tips_needed_20uL + dna_tips_needed_20uL
            self.tips_needed["p300"] = water_tips_needed_300uL + dna_tips_needed_300uL
            self.tips_needed["p1000"] = water_tips_needed_1000uL + dna_tips_needed_1000uL

            # Add required number of tip boxes to the loaded pipettes
            self.add_tip_boxes_to_pipettes()

            # Load source labware #
            ## Calculate number of water aliquots required
            total_water_required = sum(water_to_add)
            aliquots_required = math.ceil(total_water_required/self.water_aliquot_volume)
            ## Load required water source labware
            water_source_labware, water_source_locations = _OTProto.calculate_and_load_labware(self._protocol, self.water_source_labware_type, aliquots_required, custom_labware_dir = self.custom_labware_dir)

            ## Load DNA Source Labware and get locations
            DNA_labware = _OTProto.load_labware(self._protocol, self.dna_source_type, custom_labware_dir = self.custom_labware_dir, label = "DNA Source Labware")
            DNA_Locations = []
            # If DNA source wells were not specified, then choose wells to use
            if self.dna_source_wells == None:
                if len(DNA_labware.wells_by_name()) < len(self.dna):
                    raise _BMS.BMSTemplateError("Cannot currently use two source labware.")
                self.dna_source_wells = []
                for dna, source_well in zip(self.dna, DNA_labware.wells_by_name()):
                    self.dna_source_wells.append(source_well)

            for dna_well in self.dna_source_wells:
                DNA_Locations.append(DNA_labware.wells_by_name()[dna_well])

            # Load destination labware #
            destination_labware = _OTProto.load_labware(self._protocol, self._destination_labware_type, custom_labware_dir = self.custom_labware_dir, label = "Destination Labware")
            destination_locations = []
            for destination_well in self._destination_labware_wells:
                destination_locations.append(destination_labware.wells_by_name()[destination_well])

            if not _OTProto.get_p20 == None:
                self._protocol.pause("This protocol uses {} 20 uL tip boxes".format(_OTProto.tip_racks_needed(self.tips_needed["p20"], self.starting_tips["p20"])))
            if not _OTProto.get_p300 == None:
                self._protocol.pause("This protocol uses {} 300 uL tip boxes".format(_OTProto.tip_racks_needed(self.tips_needed["p300"], self.starting_tips["p300"])))
            if not _OTProto.get_p1000 == None:
                self._protocol.pause("This protocol uses {} 1000 uL tip boxes".format(_OTProto.tip_racks_needed(self.tips_needed["p1000"], self.starting_tips["p1000"])))

            self._protocol.pause("This protocol uses {} aliquots of {} uL water, located at {}".format(len(water_source_locations), self.water_aliquot_volume, water_source_locations))

            for dna, location, volume in zip(self.dna, DNA_Locations, dna_to_add):
                self._protocol.pause("Place DNA sample {} at {}. {} uL will be used".format(dna, location, volume))
            # Liquid handling begins #
            ## Dispense water into DNA source wells and mix
            _OTProto.dispense_from_aliquots(self._protocol, water_to_add, water_source_locations, destination_locations, new_tip = True)

            ## Add DNA to water
            _OTProto.transfer_liquids(self._protocol, dna_to_add, DNA_Locations, destination_locations, new_tip = True, mix_after = (5,"transfer_volume"))



import BiomationScripter as _BMS
import BiomationScripter.EchoProto as _EchoProto
import math
from typing import List, NewType, Dict, Tuple, Union
from decimal import Decimal


class Template(_EchoProto.EchoProto_Template):
    def __init__(
        self,
        Source_Colours: List[str],
        Final_Volume: float,
        Mixing_Ratios: List[str],
        Permutations: bool = False,
        Merge: bool = True,
        **kwargs # This will make the superclass arguments available to `Colour_Mixing` as keyword arguments
    ):
        super().__init__(**kwargs) # This passes the keyword arguments to the superclass

        #######################
        # User Defined Values #
        #######################
        self.source_colours = Source_Colours
        self.final_volume = Final_Volume # uL
        self.mixing_ratios = Mixing_Ratios
        self.permutations = Permutations
        self.merge = Merge

    def run(self):

        # Set up an empty list in which the colour mixtures required will be added
        Colour_Mixtures = []

        # Iterate through the list of source colours provided to create the list of mixtures.
        for colour_1 in self.source_colours:
            for colour_2 in self.source_colours:
                # Ignore situations where colour_1 and colour_2 are the same
                if colour_1 == colour_2:
                    continue
                # Unless permutations has been set to `True`, ignore situations where...
                # ...the same colours have already been mixed, just in a different order
                elif not self.permutations and [colour_2, colour_1] in Colour_Mixtures:
                    continue
                else:
                    # Add the two colours to the list of mixtures to prepare
                    Colour_Mixtures.append([colour_1, colour_2])

        # Here, we'll print to OUT all of the mixtures which will be prepared
        for c in Colour_Mixtures:
            print(c)

        # Determine how may different mixtures will be prepared
        Number_Of_Mixtures = len(Colour_Mixtures) * len(self.mixing_ratios)

        # Use BMS.Create_Labware_Needed to ensure enough destination plates are available
        Extra_Destination_Plates_Required = _BMS.Create_Labware_Needed(
            Labware_Format = self.destination_plate_layouts[0],
            N_Wells_Needed = Number_Of_Mixtures,
            N_Wells_Available = "All",
            Return_Original_Layout = False
        )

        # If any extra plates were created, add them using the `add_destination_layout` method defined by the superclass
        for plate_layout in Extra_Destination_Plates_Required:
            self.add_destination_layout(plate_layout)

        # Add content to the destination plate(s)

        # A counter is used to iterate through the destination plates (if more than one)
        Destination_Plate_Index = 0

        for mixture in Colour_Mixtures:
            # Get the current destination plate
            Destination_Plate = self.destination_plate_layouts[Destination_Plate_Index]

            # Get the colours
            colour_1 = mixture[0]
            colour_2 = mixture[1]

            # For every ratio defined by the user
            for ratio in self.mixing_ratios:
                # Get the ratio for each colour
                colour_1_ratio = float(ratio.split(":")[0])
                colour_2_ratio = float(ratio.split(":")[1])

                # Get the volumes to add for each colour
                colour_1_volume = (colour_1_ratio/(colour_1_ratio + colour_2_ratio)) * self.final_volume
                colour_2_volume = (colour_2_ratio/(colour_1_ratio + colour_2_ratio)) * self.final_volume

                # Get the next empty well in the destination plate
                well = Destination_Plate.get_next_empty_well()
                # If there are no empty wells left, iterate to the next plate and try again
                if not well:
                    Destination_Plate_Index += 1
                    Destination_Plate = self.destination_plate_layouts[Destination_Plate_Index]
                    well = Destination_Plate.get_next_empty_well()

                # Add content to the plate
                Destination_Plate.add_content(
                    Well = well,
                    Reagent = colour_1,
                    Volume = colour_1_volume
                )
                Destination_Plate.add_content(
                    Well = well,
                    Reagent = colour_2,
                    Volume = colour_2_volume
                )

        self.create_picklists()



import os
import math as _math
import random
import warnings
from typing import Dict, List, Literal, Union, Tuple

try:
    import pandas as pd
except ImportError:
    warnings.warn("No pandas, continue at your own zoological peril 🐼")

# Exception classes #
class BiomationError(Exception):
    """A general exception for BiomationScripter"""
    pass

class NegativeVolumeError(BiomationError):
    pass


class RobotConfigurationError(BiomationError):
    pass


class BMSTemplateError(BiomationError):
    pass


class LabwareError(BiomationError):
    pass


class OutOFSourceMaterial(BiomationError):
    pass


class DoEError(BiomationError):
    pass


class TransferError(BiomationError):
    pass


# Function to pass template modules
def get_template_module(module_name: Literal["OTProto", "EchoProto"]):
    if module_name == "OTProto":
        import BiomationScripter.OTProto.Templates as OTProto_Templates

        return OTProto_Templates

    elif module_name == "EchoProto":
        import BiomationScripter.EchoProto.Templates as EchoProto_Templates

        return EchoProto_Templates

    raise ValueError(f"invald module_name ({module_name!r}) passed to get_template_module")


#####################
# Classes


class DoE_Material:
    def __init__(self, ID, Material_Type, Factors=None, Factor_Values=None):
        if (Factors and not Factor_Values) or (not Factors and Factor_Values):
            raise DoEError(
                f"Error creating source type {Material_Type} {ID}"
                " - factor names or factor values were specified without the other."
            )

        self.id = ID
        self.material_type = Material_Type
        self.factors = {}

        if Factors:
            for factor, value in zip(Factors, Factor_Values):
                self.factors[factor] = value

    def get_factor_value(self, Factor):
        return self.factors[Factor]


class DoE_Intermediate:
    def __init__(
        self,
        ID,
        Intermediate_Type,
        Component_Types,
        Component_IDs,
        Component_Amount_Types,
        Component_Amount_Values,
    ):
        self.id = ID
        self.intermediate_type = Intermediate_Type
        self.components = {}

        for component_type, id, amount_type, amount_value in zip(
            Component_Types, Component_IDs, Component_Amount_Types, Component_Amount_Values
        ):
            self.components[component_type] = [id, amount_type, amount_value]

    def get_component_id(self, Component):
        return self.components[Component][0]

    def get_component_amount_info(self, Component):
        return self.components[Component][1:]


class DoE_Experiment:
    def __init__(self, Name, DoE_File):
        self.name = Name
        self.doe_file = DoE_File
        self.materials = {}
        self.intermediates = {}

        # Open the CSV file containing the DoE run data
        file = open(DoE_File, "r")
        # Store the DoE data from the file as a list
        data = []
        for row in file:
            data.append(row)

        # Close the file
        file.close()

        # Remove any empty lines imported
        if "" in data:
            data.remove("")

        # Get the factor names
        self.factors = data[0].replace("\n", "").split(",")

        # store run data using the DoE_Run_Data class
        self.runs = []
        ID = 0
        for run in data[1:]:
            # remove new line symbols if present, and convert string to a list
            run = run.replace("\n", "").split(",")
            if "" in run:
                raise ValueError(
                    f"Missing value found in run {ID} for factor {self.factors[run.index('')]}; "
                    "check that all values are present and objectives are not present in the file."
                )
            self.runs.append(DoE_Run_Data(ID, self.factors, run))
            ID += 1

    def add_material(self, Material):
        self.materials[Material.id] = Material

    def get_material(self, Material_Name):
        return self.materials[Material_Name]

    def add_intermediate(self, Intermediate):
        self.intermediates[Intermediate.id] = Intermediate

    def get_intermediate(self, Intermediate_Name):
        return self.intermediates[Intermediate_Name]

    def get_run(self, ID):
        for run in self.runs:
            if run.id == ID:
                return run

    def batch_by_factor_value(self, Factor, Value):
        batched_experiment = DoE_Experiment(
            "{}-Batched-{}-{}".format(self.name, Factor, Value), self.doe_file
        )
        runs = batched_experiment.runs.copy()
        for run in runs:
            if not run.get_factor_value(Factor) == Value:
                batched_experiment.runs.remove(run)
        return batched_experiment

    def get_all_values(self, Name):
        Values = []
        for run in self.runs:
            Values.append(run.get_value_by_name(Name))

        return Values


class DoE_Run_Data:
    def __init__(self, ID, Factors, Values):
        self.id = ID
        self.source_materials = {}
        self.intermediates = {}
        self.run_data = {}
        for factor, value in zip(Factors, Values):
            self.run_data[factor] = value

    def get_factor_value(self, Factor):
        return self.run_data[Factor]

    def add_factor(self, Factor, Value):
        self.run_data[Factor] = Value

    def specify_source_material(self, Material, Value):
        self.source_materials[Material] = Value

    def get_source_material_value(self, Source_Material):
        return self.source_materials[Source_Material]

    def specify_intermediate(self, Intermediate, Value):
        self.intermediates[Intermediate] = Value

    def get_intermediate_value(self, Intermediate):
        return self.intermediates[Intermediate]

    def get_value_by_name(self, Name):
        found = False
        Value = None
        if not found:
            try:
                Value = self.get_factor_value(Name)
                found = True
            except KeyError:
                pass
        if not found:
            try:
                Value = self.get_source_material_value(Name)
                found = True
            except KeyError:
                pass
        if not found:
            try:
                Value = self.get_intermediate_value(Name)
                found = True
            except KeyError:
                pass
        if not found:
            raise DoEError("Cannot find {}.".format(Name))
        else:
            return Value


class Labware_Content:
    """This class is used by `Labware_Layout` to store content information."""

    def __init__(self, Name: str, Volume: Union[float, int], Liquid_Class: Union[str, None] = None):
        self.name = Name
        self.volume = Volume
        self.liquid_class = Liquid_Class

    def get_info(self):
        return [self.name, self.volume, self.liquid_class]


class Assembly:
    """This class is used to store basic information about a DNA assembly."""

    def __init__(self, Name: str, Backbone: str, Parts: List[str]):
        self.name = Name
        self.backbone = Backbone
        self.parts = Parts


class Labware_Layout:
    """This class is used to store information, such as number of wells and content, about labware which can later be retrieved.

    The BMS.Labware_Layout class is intended to be universal within BiomationScripter, and not specific to any particular automation equipment.
    Labware_Layout objects are separate from, but sometimes related to, any object which is meant to represent a physical instance of a piece of labware.
    """

    def __init__(self, Name: str, Type: str):
        self.name: str = Name
        self.type: str = Type
        self.rows: int = None
        self.columns: int = None
        self.content: Dict[str, List[Labware_Content]] = {}
        self.available_wells: List[str] = None
        self.empty_wells: List[str] = None
        self.well_labels: Dict[str, str] = {}

    def get_total_volume_of_liquid(self, Liquid):
        wells = self.get_wells_containing_liquid(Liquid)
        total_volume = 0
        for well in wells:
            total_volume += self.get_volume_of_liquid_in_well(Liquid, well)

        return total_volume

    def define_format(self, Rows: int, Columns: int):
        self.rows = Rows
        self.columns = Columns

    def get_format(self):
        return (self.rows, self.columns)

    def set_available_wells(
        self, Well_Range=None, Use_Outer_Wells=True, Direction="Horizontal", Box=False
    ):
        self.available_wells = self.get_well_range(Well_Range, Use_Outer_Wells, Direction, Box)
        self.empty_wells = self.available_wells.copy()

    def get_available_wells(self):
        return self.available_wells

    def get_well_range(
        self, Well_Range=None, Use_Outer_Wells=True, Direction="Horizontal", Box=False
    ):
        if not Direction == "Horizontal" and not Direction == "Vertical":
            raise ValueError("`Direction` must be either 'Horizontal' or 'Vertical'")

        n_rows = self.rows
        n_cols = self.columns
        if Well_Range is None:
            Well_Range = "A1:{}{}".format(chr(64 + n_rows), n_cols)

        wells = well_range(Well_Range, Labware_Format=self, Direction=Direction, Box=Box)

        if not Use_Outer_Wells:
            plate_first_row = "A"
            plate_last_row = chr(64 + n_rows)
            plate_first_col = 1
            plate_last_col = n_cols

            temp_wells = []
            for w in wells:
                if (
                    (plate_first_row in w)
                    or (plate_last_row in w)
                    or (str(plate_first_col) == w[1:])
                    or (str(plate_last_col) == w[1:])
                ):
                    continue
                else:
                    temp_wells.append(w)
            wells = temp_wells

        return wells

    def check_well(self, Well):
        Row = Well[0]
        Column = int(Well[1:])
        InRange = None
        if Row > chr(64 + self.rows):
            InRange = False
        elif Column > self.columns:
            InRange = False
        else:
            InRange = True
        return InRange

    def clone_format(self, New_Name):
        # Does not copy content
        name = New_Name
        type = self.type
        rows = self.rows
        columns = self.columns
        plate_copy = Labware_Layout(name, type)
        plate_copy.define_format(rows, columns)
        return plate_copy

    def bulk_add_content(self, Wells, Reagents, Volumes, Liquid_Classes = None):
        if type(Volumes) is int or type(Volumes) is float:
            Volumes = [Volumes] * len(Wells)
        if Liquid_Classes is None:
            for well, reagent, volume in zip(Wells, Reagents, Volumes):
                self.add_content(well, reagent, volume)
        else:
            for well, reagent, volume, liquid_class in zip(Wells, Reagents, Volumes, Liquid_Classes):
                self.add_content(well, reagent, volume, liquid_class)

    def add_content(self, Well, Reagent, Volume, Liquid_Class=None):
        if Volume < 0:
            raise NegativeVolumeError

        if self.available_wells and ":" not in Well and not type(Well) == list:
            if Well not in self.available_wells:
                raise LabwareError(
                    f"Available wells are specified, but well {Well} is not defined as available. "
                    "Check that the correct well has been specified. "
                    f"Available wells are: {self.available_wells}\n"
                )

        # Volume should always be uL
        if Liquid_Class is None:
            Liquid_Class = "Unknown"
        if ":" in Well:
            for w in well_range(Well, Labware_Format=self):
                self.add_content(w, Reagent, Volume, Liquid_Class)
        elif type(Well) == list:
            for well in Well:
                self.add_content(well, Reagent, Volume, Liquid_Class)
        elif Well in self.content:
            self.content[Well].append(Labware_Content(Reagent, float(Volume), Liquid_Class))
        else:
            self.content[Well] = [Labware_Content(Reagent, float(Volume), Liquid_Class)]

        if self.empty_wells and Well in self.empty_wells:
            self.empty_wells.remove(Well)

    def add_well_label(self, Well: str, Label: str):
        for well in self.well_labels:
            if self.well_labels[well] == Label:
                raise LabwareError(
                    'Label "{}" is already used as a label in {}'.format(Label, well)
                )
        self.well_labels[Well] = Label

    def get_well_content_by_label(self, Label: str):
        for well in self.well_labels:
            if self.well_labels[well] == Label:
                return self.content[well]
        return None

    def get_well_location_by_label(self, Label: str):
        for well in self.well_labels:
            if self.well_labels[well] == Label:
                return well

        return None

    def get_content(self):
        return self.content

    def get_occupied_wells(self):
        return list(self.get_content().keys())

    def get_liquids_in_well(self, Well):
        if Well not in self.get_occupied_wells():
            raise LabwareError("Well {} in labware {} contains no liquids.".format(Well, self.name))

        liquids_in_well = []
        content_in_well = self.content[Well]
        for content in content_in_well:
            liquids_in_well.append(content.name)

        return liquids_in_well

    def get_wells_containing_liquid(self, Liquid_Name):
        wells_to_return = []
        content = self.get_content()
        wells = self.get_occupied_wells()
        for well in wells:
            for liquid in content[well]:
                if liquid.name == Liquid_Name:
                    wells_to_return.append(well)
        return wells_to_return

    def clear_content(self):
        self.content = {}
        if self.available_wells:
            self.empty_wells = self.available_wells.copy()

    def clear_content_from_well(self, Well):
        del self.content[Well]

        if self.empty_wells:
            self.empty_wells.append(Well)

    def clear_liquid_in_well(self, Well, Liquid):
        well_contents = self.content[Well].copy()
        for content in well_contents:
            if Liquid == content.name:
                self.content[Well].remove(content)

        # Check if the well is now empty
        if len(self.content[Well]) == 0:
            # Clean up
            self.clear_content_from_well(Well)

    def get_next_empty_well(self):
        if not self.available_wells:
            raise LabwareError(
                "Available wells must be specified to get the next empty well. "
                f"Specify the avilable wells for {self.name} using `.set_available_wells`."
            )
        if len(self.empty_wells) == 0:
            return None
        # Iterate through all available wells, and return the first one which also exists in the empty wells attribute
        for well in self.available_wells:
            if well in self.empty_wells:
                return well

    def get_volume_of_liquid_in_well(self, Liquid, Well):
        # if not Well in self.get_occupied_wells():
        #     raise LabwareError("Well {} in labware {} contains no liquids.".format(Well, self.name))
        well_content = self.get_content()[Well]
        for content in well_content:
            if content.name == Liquid:
                return content.volume

        return 0.0

    def update_volume_in_well(self, Volume, Reagent, Well):
        if Well not in self.get_content().keys():
            raise LabwareError(
                "{} has not been previously defined. Add content to this well using the `add_content` method.".format(
                    Well
                )
            )
        well_content = self.get_content()[Well]
        for content in well_content:
            if content.name == Reagent:
                content.volume = float(Volume)

    def print(self):
        print("\033[1mInformation for " + self.name + "\033[0m")
        print("Plate Type: " + self.type)
        content = self.get_content()
        print("Well\tVolume(uL)\tLiquid Class\tReagent")
        content_return = ""
        for well in content:
            for c in content[well]:
                content_return += (
                    well + "\t" + str(c.volume) + "\t\t" + c.liquid_class + "\t\t" + c.name + "\n"
                )
                print(well + "\t" + str(c.volume) + "\t\t" + c.liquid_class + "\t\t" + c.name)
        return content_return

    def import_labware(self, filename, path="~", ext=".xlsx"):
        return self.import_plate(filename, path, ext)

    # create a dummy PlateLayout object before running this method
    def import_plate(self, filename, path="~", ext=".xlsx"):
        # check if filename contains the extension
        # NOTE: this only checks for a "." so could easily throw errors if given atypical input
        if "." in filename:
            # look at Sheet1 (the Plate Metadata)
            sheet1 = pd.read_excel(path + filename, sheet_name=0, header=None, engine="openpyxl")
            # look at Sheet2 (the Well lookup)
            sheet2 = pd.read_excel(path + filename, sheet_name=1, engine="openpyxl")
        else:
            # add extension (default = ".xlsx") to arg1
            sheet1 = pd.read_excel(
                path + filename + ext, sheet_name=0, header=None, engine="openpyxl"
            )
            sheet2 = pd.read_excel(path + filename + ext, sheet_name=1, engine="openpyxl")

        # get plate name and plate type from Sheet 1
        # TODO: throw an error message if the name and/or type are blank
        self.name = sheet1[1].iloc[0]  # column 2, row 1
        self.type = sheet1[1].iloc[1]  # column 2, row 2

        # if Current Volume column is empty (NaN), replace with Initial Volume
        sheet2["Volume (uL) - Current"].fillna(sheet2["Volume (uL) - Initial"], inplace=True)
        # if both Volume columns are empty, replace with zeroes
        # TODO: add a warning to the user to say they are empty
        sheet2["Volume (uL) - Current"].fillna(0, inplace=True)
        # TODO: add a warning that liquid class has been set to a default value
        sheet2["Calibration Type"].fillna("AQ_BP", inplace=True)
        # select subset of columns as "content"
        content = sheet2[["Well", "Name", "Volume (uL) - Current", "Calibration Type"]]
        # delete rows where the Name is blank
        content = content.dropna(subset=["Name"])
        # iterate over rows
        for row in content.itertuples(index=False):
            # add contents to PlateLayout object
            self.add_content(row[0], row[1], row[2], row[3])
        return self


class Liquids:
    """This class is used to store information about liquids and where they are stored."""

    def __init__(self):
        self.liquids = {}

    def add_liquid(self, liquid: str, labware: Union[str, Labware_Layout], source_well: str):
        """Adds a liquid to self.liquids"""
        self.liquids[liquid] = [labware, source_well]

    def get_liquids_in_labware(self, labware: Union[str, Labware_Layout]):
        """Returns all liquids within a specified labware."""
        liquids_to_return = []
        for liquid in self.liquids:
            if labware == self.get_liquid_labware(liquid):
                liquids_to_return.append(liquid)

        return liquids_to_return

    def get_liquid_by_location(self, labware, well):
        liquid_to_return = None
        for liquid in self.liquids:
            if labware == self.get_liquid_labware(liquid) and well == self.get_liquid_well(liquid):
                liquid_to_return = liquid
                break
        if liquid_to_return is None:
            raise ValueError("No liquid found in labware {} at well {}".format(labware, well))
        else:
            return liquid_to_return

    def get_liquid(self, liquid):
        return self.liquids[liquid]

    def get_liquid_labware(self, liquid):
        return self.liquids[liquid][0]

    def get_liquid_well(self, liquid):
        return self.liquids[liquid][1]

    def add_liquids_to_labware(self, liquids, labware, blocked_wells=None, well_range=None):
        if well_range:
            wells = well_range
        else:
            wells = list(labware.wells_by_name().keys())

        if blocked_wells:
            for bw in blocked_wells:
                wells.remove(bw)
        for liquid in liquids:
            self.liquids[liquid] = [labware, wells[0]]
            wells.remove(wells[0])

    def get_all_liquids(self):
        return list(self.liquids.keys())


class Mastermix:
    """This class is used by BMS.Mastermix_Maker to store information about a mastermix. It is not intended for use outside of this function."""

    def __init__(self, Name: str, Reagents: str, Wells: List[str]):
        self.name = Name
        self.reagents = Reagents
        self.wells = Wells

##########################################
# Functions


def Aliquot_Calculator(Liquid, Destination_Layouts, Aliquot_Volume, Dead_Volume):
    # Calculate total volume of liquid in all destinations provided
    total_volume = 0
    for destination_layout in Destination_Layouts:
        total_volume += destination_layout.get_total_volume_of_liquid(Liquid)

    num_aliquots = _math.ceil(total_volume / (Aliquot_Volume - Dead_Volume))

    return(num_aliquots)

def Get_Transfers_Required(Liquid, Destination_Layouts):
    transfer_volumes = []
    destination_wells = []
    destination_layouts = []

    for destination_layout in Destination_Layouts:
        wells_with_liquid = destination_layout.get_wells_containing_liquid(Liquid)

        for well in wells_with_liquid:
            transfer_volumes.append(destination_layout.get_volume_of_liquid_in_well(Liquid, well))
            destination_wells.append(well)
            destination_layouts.append(destination_layout)

    return(transfer_volumes, destination_wells, destination_layouts)

def DoE_Get_Value_From_Combined_Factor(Factor_Name, Combined_Factor):
    Value = Combined_Factor.split(Factor_Name)[1].split("(")[1].split(")")[0]

    # Try and return as a float - if a ValueError is raised, assume the value is a string and just return as is
    try:
        Value = float(Value)
    except ValueError:
        pass

    return Value

def DoE_Create_Source_Material(
    DoE_Experiment, Source_Material_Name, Factor_Names, Add_To_Runs=True
):
    # Create an empty list to store each of source material types
    # Each source material type is formed from combinations of the same factors...
    # ... with different values as specified by the DoE
    Source_Material_Types = []

    Source_Material_Component_Info = {}  # ID: list(Factor Values)

    if Factor_Names:
        # For each run in the DoE
        for run in DoE_Experiment.runs:
            # Get the values of all the specified factors
            source_material_type = []
            component_factor_values = []
            for component_factor in Factor_Names:
                component_factor_value = run.get_factor_value(component_factor)
                component_factor_values.append(component_factor_value)
                # Store the factor name-value combination as a string in a list
                source_material_type.append(
                    "{}({})".format(component_factor, component_factor_value)
                )
            # Convert the list of factor name-value combinations to a string and add to the previously created list
            Source_Material_Types.append("-".join(source_material_type))
            Source_Material_Component_Info["-".join(source_material_type)] = component_factor_values
            if Add_To_Runs:
                run.specify_source_material(Source_Material_Name, "-".join(source_material_type))
        for material_id in set(Source_Material_Types):
            Material_Object = DoE_Material(
                material_id,
                Source_Material_Name,
                Factor_Names,
                Source_Material_Component_Info[material_id],
            )
            DoE_Experiment.add_material(Material_Object)
    else:
        Source_Material_Types = [Source_Material_Name]
        Source_Material_Component_Info = {Source_Material_Name: None}
        if Add_To_Runs:
            for run in DoE_Experiment.runs:
                run.specify_source_material(Source_Material_Name, Source_Material_Name)

        Material_Object = DoE_Material(Source_Material_Name, Source_Material_Name)
        DoE_Experiment.add_material(Material_Object)

    return Source_Material_Types


def DoE_Create_Intermediate(
    DoE_Experiment,
    Intermediate_Name,
    Source_Material_Names,
    Source_Materials_Amount_Types,
    Source_Materials_Amount_Values,
    Add_To_Runs=True,
):
    # Create an empty list to store each of intermediate types
    # Each intermediate type is formed from combinations of the same source materials...
    # ... with different values as specified by the DoE
    Intermediate_Types = []

    Intermediate_Component_Info = {}

    # For each run in the DoE
    for run in DoE_Experiment.runs:
        # Get the values of all the specified source_materials
        intermediate_type = []
        component_source_material_values = []
        for component_source_material, amount_type, amount_value in zip(
            Source_Material_Names, Source_Materials_Amount_Types, Source_Materials_Amount_Values
        ):
            component_source_material_value = run.get_source_material_value(
                component_source_material
            )

            # Check to see if any factors are specified as references for amounts of components to be added
            # Different amounts of a source material being added results in extra combinations
            # e.g. an intermediate type might be composed of a buffer and a chemical
            # But the amount of chemical is determined by Factor-1, which has two values in the DoE
            # Therefore, there will need to be two unique versions of the intermediate type to account for
            # these differences

            if amount_value in DoE_Experiment.factors:
                amount_value = run.get_factor_value(amount_value)
                component_source_material = "{}[{}]".format(component_source_material, amount_value)

            component_source_material_values.append([component_source_material_value, amount_value])
            # Store the source_material name-value combination as a string in a list
            intermediate_type.append(
                "{}({})".format(component_source_material, component_source_material_value)
            )

        # Convert the list of source_material name-value combinations to a string and add to the previously created list
        Intermediate_Types.append("-".join(intermediate_type))
        Intermediate_Component_Info["-".join(intermediate_type)] = component_source_material_values
        if Add_To_Runs:
            run.specify_intermediate(Intermediate_Name, "-".join(intermediate_type))

    for intermediate_id in set(Intermediate_Types):
        component_ids = [id[0] for id in Intermediate_Component_Info[intermediate_id]]
        component_amount_values = [id[1] for id in Intermediate_Component_Info[intermediate_id]]
        Intermediate_Object = DoE_Intermediate(
            intermediate_id,
            Intermediate_Name,
            Source_Material_Names,
            component_ids,
            Source_Materials_Amount_Types,
            component_amount_values,
        )
        DoE_Experiment.add_intermediate(Intermediate_Object)

    return Intermediate_Types


def Reagent_Finder(Reagents: List[str], Directories: List[str]):
    """Searches a list of directories containing labware layout files for a specified reagent.

    This function will search in the directories listed for files which appear to be BMS labware layout files.
    Any labware layout files found will then be searched for the specified reagents.
    The name of all reagents are then printed to OUT, along with the name of the labware layout they appear in and the well(s) they occupy.
    """

    # Get all labware layouts in the specified directories
    Files = []
    for directory in Directories:
        Files.append(
            [
                file
                for file in os.listdir(directory + "/")
                if os.path.isfile(os.path.join(directory + "/", file)) and "~$" not in file
            ]
        )

    Layouts = []
    for files, directory in zip(Files, Directories):
        for file in files:
            try:
                Layouts.append(Import_Labware_Layout(Filename=file, path=directory + "/"))
            except:
                continue

    # Check if each reagent is in a labware file
    for reagent in Reagents:
        print("\n{}:".format(reagent))
        for layout in Layouts:
            if len(layout.get_wells_containing_liquid(reagent)) > 0:
                print("> {}: {}".format(layout.name, layout.get_wells_containing_liquid(reagent)))


def Import_Labware_Layout(Filename: str, path: str = "~", ext: str = ".xlsx"):
    """Imports an Excel file with a standard layout and converts it to a BiomationScripter.Labware_Layout object."""


    labware_layout = Labware_Layout("name", "type")
    labware_layout.import_labware(Filename, path=path, ext=ext)
    return labware_layout


def Create_Labware_Needed(
    Labware_Format: Labware_Layout, N_Wells_Needed: int, N_Wells_Available: Union[int, Literal["All"]] = "All", Return_Original_Layout: bool = True
):
    """calculates how many labware of a certain type are required for a protocol and returns a list of Labware_Layout objects.

    Determines number of labware by the total number of wells required (N_Wells_Needed) and the number of wells available per labware (N_Wells_Available).
    The labware created is based on the Labware_Layout provided to Labware_Format.
    Any content already in the template Labware_Layout is not copied to new labware created.
    """

    if not type(N_Wells_Available) is int:
        if N_Wells_Available == "All":
            N_Wells_Available = Labware_Format.rows * Labware_Format.columns
        else:
            raise ValueError("`N_Wells_Available` should either be an integer, or 'All'.")

    N_Plates_Needed = _math.ceil(N_Wells_Needed / N_Wells_Available)
    if Return_Original_Layout is True:
        Plates = [Labware_Format]
    elif Return_Original_Layout is False:
        Plates = []
    for plate_n in range(1, N_Plates_Needed):
        Plate_Name = Labware_Format.name + str(plate_n)
        Plates.append(Labware_Format.clone_format(Plate_Name))
    return Plates


def well_range(Wells: str, Labware_Format: Union[Labware_Layout, Tuple[int, int]] = None, Direction: Union[Literal["Horizontal", "Vertical"]] = "Horizontal", Box: bool = False):
    """Returns a list of wells based on a specified well range and direction.

    Wells are always in the format of row followed by column, where row is a letter and column is an integer (e.g. A1, D6, C12, B7, etc.).
    The Direction argument determines the order in which wells are counted.
    The Box argument determines whether the well range has a box-like shape.
    If Box = False, then a Labware_Format must be specified
    """

    if not Direction == "Horizontal" and not Direction == "Vertical":
        raise ValueError("`Direction` must be either 'Horizontal' or 'Vertical'")

    if not Labware_Format and not Box:
        raise LabwareError("`Box` can only be `False` when `Labware_Format` is specified")

    if not Labware_Format or Box:
        Well = Wells
        first, last = Well.split(":")
        firstL = first[0]
        lastL = last[0]
        firstN = int(first[1 : len(first)])
        lastN = int(last[1 : len(last)])
        wells = []
        if Direction == "Horizontal":
            for L in _Lrange(firstL, lastL):
                for N in range(firstN, lastN + 1):
                    wells.append(L + str(N))
            return wells
        elif Direction == "Vertical":
            for N in range(firstN, lastN + 1):
                for L in _Lrange(firstL, lastL):
                    wells.append(L + str(N))
            return wells

    else:
        # Get the end row and end column number for the labware being used
        if type(Labware_Format) is list:
            if not len(Labware_Format) == 2:
                raise ValueError(
                    "Labware_Format argument MUST either be a BiomationScripter.Labware_Layout "
                    "object, or a list specifying number of rows and number of columns (e.g. "
                    "[8, 12])."
                )
            end_row, end_col = Labware_Format
        else:
            Labware_Format = [Labware_Format.rows, Labware_Format.columns]

        end_row, end_col = Labware_Format
        # Get the first and last wells in the specified well range
        first_well, last_well = Wells.split(":")

        # Split the first and last wells up into rows and columns
        first_row = first_well[0]
        last_row = last_well[0]
        first_col = int(first_well[1:])
        last_col = int(last_well[1:])

        # Check if first and last wells are in range
        if (
            (_Labware_Row_To_Index(first_row) > end_row - 1)
            or (_Labware_Row_To_Index(last_row) > end_row - 1)
            or (first_col > end_col)
            or (last_col > end_col)
        ):
            raise ValueError("Wells are not in range of specified format")

        wells = []

        if Direction == "Horizontal":
            rows = []
            for r in _Lrange(first_row, last_row):
                rows.append(r)

            first_col_of_row = first_col

            for row in rows:
                for col in range(first_col_of_row, end_col + 1):
                    well = "{}{}".format(row, col)
                    wells.append(well)
                    if col == end_col:
                        first_col_of_row = 1
                    if well == last_well:
                        return wells

        elif Direction == "Vertical":
            rows = []
            for r in _Lrange("A", chr(64 + end_row)):
                rows.append(r)

            first_row_of_col_index = _Labware_Row_To_Index(first_row)

            for col in range(first_col, last_col + 1):
                for row in rows[first_row_of_col_index:]:
                    well = "{}{}".format(row, col)
                    wells.append(well)
                    if row == rows[-1]:
                        first_row_of_col_index = 0
                    if well == last_well:
                        return wells


def Group_Locations(Locations, Group_Populations):
    Grouped_Locations = []
    start_index = 0
    end_index = 0
    for group_pop in Group_Populations:
        end_index += group_pop
        Grouped_Locations.append(Locations[start_index:end_index])
        start_index += group_pop
    return Grouped_Locations


def Mastermix_Maker(
    Destination_Layouts: List[Labware_Layout],
    Mastermix_Layout: Labware_Layout,
    Maximum_Mastermix_Volume: float,
    Min_Transfer_Volume: float,
    Extra_Reactions: float,
    Excluded_Reagents: List[str] = [],
    Excluded_Combinations: List[List[str]] = [],
    Preferential_Reagents: List[str] = [],
    Seed: Union[int, None] = None,
):

    """Generates mastermixes based on source materials in a list of destination Labware_Layout objects, and other user-defined parameters.

    Mastermixes are generated by finding common source reagents across destination wells and grouping them together in the minimal amount of mastermixes.
    Parameters supplied as arguments can be used to influence compositions of the mastermixes.
    To help ensure many combinations are attempted, there is some randomness within the function.
    This randomness can be removed by supplying a specific seed, which ensures that the same mastermixes are generated each time.
    """

    Solved = False  # Check whether a solution has been found

    First_Attempt = True  # Determine whether this is the first attempt to find a solution

    # Whilst there is no solution
    print("\nDetermining mastermixes, this may take a while...\n")
    while not Solved:

        try:

            # Get all unique reagents in the destination layout
            All_Reagents = set()

            for Destination_Layout in Destination_Layouts:
                for well in Destination_Layout.get_content():
                    for reagent in Destination_Layout.get_liquids_in_well(well):
                        All_Reagents.add(reagent)

            # For each reagent at each volume/well, get the wells it appears in
            # e.g., 20 uL water might be in wells A1, A2, A3 and 30 uL water might be in wells A4, A5, A6

            Wells_By_Reagent_ID = {}

            for reagent in All_Reagents:
                # get wells containing the current reagent (grouped by destination layout)
                wells_with_reagent_by_destination = [
                    Destination_Layout.get_wells_containing_liquid(reagent)
                    for Destination_Layout in Destination_Layouts
                ]

                # combine grouped wells into one list, and add the destination layout index to the
                # well designation (INDEX_WELL)
                wells_with_reagent = []
                for wells, dest_index in zip(
                    wells_with_reagent_by_destination,
                    range(0, len(wells_with_reagent_by_destination)),
                ):
                    wells_with_reagent += ["{}_{}".format(dest_index, well) for well in wells]

                # Group the identified wells based on the reagent's volume/well
                for well in wells_with_reagent:
                    # Get the volume in the well
                    well_volume = Destination_Layouts[
                        _get_well_layout_index(well)
                    ].get_volume_of_liquid_in_well(reagent, well.split("_")[1])
                    # Create a new reagent ID which includes the volume/well
                    reagent_id = "{}_vol_{}".format(reagent, well_volume)
                    # Add the well to the relevant reagent id (create the entry if needed)
                    if reagent_id in Wells_By_Reagent_ID.keys():
                        Wells_By_Reagent_ID[reagent_id].add(well)
                    else:
                        Wells_By_Reagent_ID[reagent_id] = set([well])

            # a set of all volume/well values
            vols_per_well = set()
            # Get the reagent id keys in alphabetical order - needed to ensure consistency when doing random shuffles
            reagent_id_keys = list(Wells_By_Reagent_ID.keys())
            reagent_id_keys.sort()

            for reagent_id in reagent_id_keys:
                vols_per_well.add(_get_vol_per_well(reagent_id))

            vols_per_well = list(vols_per_well)
            vols_per_well.sort()

            # Get a list of reagent IDs ordered by volume/well (ascending)
            Ordered_Reagent_IDs = []
            for volume in vols_per_well:
                for reagent_id in reagent_id_keys:
                    if _get_vol_per_well(reagent_id) == volume:
                        Ordered_Reagent_IDs.append(reagent_id)

            # Removed excluded reagents (i.e. those not to be used in mastermixes)

            for (
                reagent
            ) in Ordered_Reagent_IDs.copy():  # copy to make sure the iterating works as expected
                for excluded_reagent in Excluded_Reagents:
                    if excluded_reagent in reagent:
                        Ordered_Reagent_IDs.remove(reagent)

            #####################################

            # Check for linked reagents (where reagent ids share the same well set)

            reagents_checked = []  # Used to ensure combinations, NOT permutations
            Linked_Reagents = []

            for reagent_id in Ordered_Reagent_IDs:
                linked_reagents = [reagent_id]  # temp list for inside loop
                reagent_well_set = Wells_By_Reagent_ID[
                    reagent_id
                ]  # well set for the current reagent
                for candidate_reagent in Ordered_Reagent_IDs:
                    if candidate_reagent == reagent_id:
                        # Don't link a reagent with itself
                        continue
                    elif candidate_reagent in reagents_checked:
                        # Ignore already checked reagents - avoids permutations
                        continue
                    elif reagent_well_set == Wells_By_Reagent_ID[candidate_reagent]:
                        # If the reagent well sets are identical, then link the reagents
                        linked_reagents.append(candidate_reagent)
                        # Make sure that the linked reagents aren't excluded combinations
                        for excluded_combo in Excluded_Combinations:
                            if (
                                candidate_reagent.split("_vol_")[0] in excluded_combo
                                and reagent_id.split("_vol_")[0] in excluded_combo
                            ):
                                linked_reagents.remove(
                                    candidate_reagent
                                )  # remove it if it is an excluded combo

                # Reagent volumes will be scaled up to ensure that the mastermix can service the correct number of wells,
                ## and to also ensure that the lowest volume reagent can be added with a transfer volume above the min transfer vol
                # Check that this scale up will not put the total mastermix volume above the total well capacity if these linked reagents were used
                lowest_vol_reag = [
                    reag
                    for reag in linked_reagents
                    if _get_vol_per_well(reag)
                    == min([_get_vol_per_well(reag) for reag in linked_reagents])
                ][0]
                if (
                    _get_vol_per_well(lowest_vol_reag)
                    * (len(Wells_By_Reagent_ID[lowest_vol_reag]) + Extra_Reactions)
                    >= Min_Transfer_Volume
                ):
                    potential_mm_reaction_num = (
                        len(Wells_By_Reagent_ID[lowest_vol_reag]) + Extra_Reactions
                    )
                else:
                    potential_mm_reaction_num = Min_Transfer_Volume / _get_vol_per_well(
                        lowest_vol_reag
                    )
                # If the sum of all mm reagents multiplied by the number of required reagents would be more than the max mastermix vol allowed
                ## then remove the highest vol reagent until the vol is below the max volume allowed
                total_potential_mastermix_volume = sum(
                    [
                        _get_vol_per_well(reag) * potential_mm_reaction_num
                        for reag in linked_reagents
                    ]
                )
                while total_potential_mastermix_volume > Maximum_Mastermix_Volume:
                    linked_reagents.remove(
                        [
                            reag
                            for reag in linked_reagents
                            if _get_vol_per_well(reag)
                            == max([_get_vol_per_well(reag) for reag in linked_reagents])
                        ][0]
                    )
                    total_potential_mastermix_volume = sum(
                        [
                            _get_vol_per_well(reag) * potential_mm_reaction_num
                            for reag in linked_reagents
                        ]
                    )

                # Check if the current reagent was linked with any other reagent(s)
                if len(linked_reagents) == 1:
                    pass
                else:
                    Linked_Reagents.append(linked_reagents)

                # Mark that the reagent has been checked
                reagents_checked.append(reagent_id)

            # print(Linked_Reagents)

            ######################################

            # Create the mastermixes

            No_Solution = False

            if Seed:
                random.seed(Seed)
                random.shuffle(Ordered_Reagent_IDs)

            Mastermixes = []
            for reagent_id in Ordered_Reagent_IDs:

                # Check if the reagent's volume/well is below the trasfer threshold
                if _get_vol_per_well(reagent_id) < Min_Transfer_Volume:
                    # print("\n")
                    mastermix_reagents = [reagent_id]  # list to hold the mastermix components
                    mastermix_per_well = _get_vol_per_well(
                        reagent_id
                    )  # to track the volume of MM to add per well
                    current_reagent_well_set = Wells_By_Reagent_ID[
                        reagent_id
                    ].copy()  # set of wells which need to be statisifed by the MMs

                    # Check if the reagent's well set is empty
                    ## This can occur if the reagent has been satisified by other mastermixes
                    if len(current_reagent_well_set) == 0:
                        continue

                    # print("start", mastermix_reagents)

                    ##########################################
                    # Check if there are any linked reagents #
                    ##########################################
                    for linked_reagents in Linked_Reagents:
                        if reagent_id in linked_reagents:
                            # For any linked reagents found...
                            for linked_reagent in linked_reagents:
                                # Skip the current reagent in the list so it isn't added to itself
                                if linked_reagent == reagent_id:
                                    continue
                                # Check that the linked reagent hasn't already been used up in other mastermixes
                                elif len(Wells_By_Reagent_ID[linked_reagent]) == 0:
                                    continue
                                else:
                                    mastermix_reagents.append(linked_reagent)
                                    mastermix_per_well += _get_vol_per_well(linked_reagent)
                    # Check if the MM per well is above the threshold
                    if mastermix_per_well >= Min_Transfer_Volume:
                        # print("Reagent:", reagent_id)
                        # print("linked mm reag", mastermix_reagents)
                        # If it is, then create the mastermix
                        mastermix_name = ":".join(mastermix_reagents)
                        # Check if any linked reagents have already been used
                        if False in [
                            current_reagent_well_set == Wells_By_Reagent_ID[mm_reag]
                            for mm_reag in mastermix_reagents
                        ]:
                            # If the well sets are not equal, then just skip to the next code block (essentially means that the simple solution failed...)
                            # Reset the variables first...
                            mastermix_reagents = [
                                reagent_id
                            ]  # list to hold the mastermix components
                            mastermix_per_well = _get_vol_per_well(
                                reagent_id
                            )  # to track the volume of MM to add per well
                            current_reagent_well_set = Wells_By_Reagent_ID[
                                reagent_id
                            ].copy()  # set of wells which need to be statisifed by the MMs
                            pass
                        else:
                            mastermix_well_set = current_reagent_well_set.copy()
                            Mastermixes.append(
                                Mastermix(
                                    mastermix_name,
                                    mastermix_reagents.copy(),
                                    mastermix_well_set.copy(),
                                )
                            )
                            # Remove the used wells from the reagent IDs
                            for used_reagent in mastermix_reagents:
                                # print("Current {} well set: {}".format(used_reagent, Wells_By_Reagent_ID[used_reagent]))
                                Wells_By_Reagent_ID[used_reagent] -= mastermix_well_set
                                # print("Updated {} well set: {}".format(used_reagent, Wells_By_Reagent_ID[used_reagent]))
                            # print("---")

                            # Then continue to the next reagent in need of a mastermix
                            continue
                    else:
                        # If not, pass to the next code block to add some more components
                        pass

                    ###############################################################
                    # Add components until the MM per well is above the threshold #
                    ###############################################################
                    split = (
                        []
                    )  # This is used to keep track of things if multiple MMs are needed for a reagent
                    ## split is a list of sets: split = [ [wells: set, mm_reagents: set] ]
                    split_active = (
                        False  # True when there are still wells that need to be dealt with
                    )
                    ## split variables hold wells that are not currently active
                    ## mastermix_ variables hold currently active wells (actively trying to satisify)

                    # Keep looping until the MM per well is above the threshold and all wells have been satisified
                    while (
                        mastermix_per_well < Min_Transfer_Volume
                        or len(current_reagent_well_set) > 0
                        or split_active
                    ):
                        # Start by identifying the reagent with the most wells in common
                        ## (if there is a tie, the reagent with lowest vol/well is selected)
                        common_wells = set()
                        chosen_component = ""
                        chosen_components = []
                        preferential_selected = False
                        for candidate_reagent in Ordered_Reagent_IDs:
                            # Make sure that the linked reagents aren't excluded combinations
                            exclude = False
                            for excluded_combo in Excluded_Combinations:
                                if (
                                    candidate_reagent.split("_vol_")[0] in excluded_combo
                                    and reagent_id.split("_vol_")[0] in excluded_combo
                                ):
                                    exclude = True
                                    break
                            if exclude:
                                continue

                            # Make sure that adding the reagent wouldn't put the mastermix volume over the max allowed
                            # Reagent volumes will be scaled up to ensure that the mastermix can service the correct number of wells,
                            ## and to also ensure that the lowest volume reagent can be added with a transfer volume above the min transfer vol
                            # Check that this scale up will not put the total mastermix volume above the total well capacity if these linked reagents were used
                            # Only bother checking if the candidate reagent shares wells with the current reagent
                            if (
                                len(
                                    current_reagent_well_set.intersection(
                                        Wells_By_Reagent_ID[candidate_reagent]
                                    )
                                )
                                > 0
                            ):
                                lowest_vol_reag = [
                                    reag
                                    for reag in mastermix_reagents + [candidate_reagent]
                                    if _get_vol_per_well(reag)
                                    == min(
                                        [
                                            _get_vol_per_well(reag)
                                            for reag in mastermix_reagents + [candidate_reagent]
                                        ]
                                    )
                                ][0]
                                if (
                                    _get_vol_per_well(lowest_vol_reag)
                                    * (len(Wells_By_Reagent_ID[lowest_vol_reag]) + Extra_Reactions)
                                    >= Min_Transfer_Volume
                                ):
                                    potential_mm_reaction_num = (
                                        len(Wells_By_Reagent_ID[lowest_vol_reag]) + Extra_Reactions
                                    )
                                    # If the sum of all mm reagents multiplied by the number of required reagents would be more than the max mastermix vol allowed, then continue
                                    total_potential_mastermix_volume = sum(
                                        [
                                            _get_vol_per_well(reag) * potential_mm_reaction_num
                                            for reag in mastermix_reagents + [candidate_reagent]
                                        ]
                                    )
                                    if total_potential_mastermix_volume > Maximum_Mastermix_Volume:
                                        # print(total_potential_mastermix_volume, Maximum_Mastermix_Volume, reagent_id, candidate_reagent)
                                        continue
                                else:
                                    potential_mm_reaction_num = (
                                        Min_Transfer_Volume / _get_vol_per_well(lowest_vol_reag)
                                    )
                                    # before continuing, if the number of reactions required by the lowest vol reagent will always make the candidate reagent above the max
                                    ## vol allowed, then mark the reagents as excluded combos to ensure that they don't keep getting checked
                                    if (
                                        _get_vol_per_well(candidate_reagent)
                                        * potential_mm_reaction_num
                                        > Maximum_Mastermix_Volume
                                    ):
                                        Excluded_Combinations.append(
                                            [candidate_reagent, lowest_vol_reag]
                                        )
                                        continue
                                    # If the sum of all mm reagents multiplied by the number of required reagents would be more than the max mastermix vol allowed, then continue
                                    total_potential_mastermix_volume = sum(
                                        [
                                            _get_vol_per_well(reag) * potential_mm_reaction_num
                                            for reag in mastermix_reagents + [candidate_reagent]
                                        ]
                                    )
                                    if total_potential_mastermix_volume > Maximum_Mastermix_Volume:
                                        continue

                            if candidate_reagent == reagent_id:
                                # don't check reagents with theirselves
                                continue
                            elif candidate_reagent in mastermix_reagents:
                                # Don't add duplicate reagents
                                continue

                            elif candidate_reagent.split("_vol_")[0] in Preferential_Reagents:
                                preferential_selected = True
                                # If the candidate is a preferential reagent, choose it over non-preferential reagents
                                if len(
                                    current_reagent_well_set.intersection(
                                        Wells_By_Reagent_ID[candidate_reagent]
                                    )
                                ) > len(common_wells):
                                    # If so, store this as the current best option (may get overwritten later)
                                    common_wells = current_reagent_well_set.intersection(
                                        Wells_By_Reagent_ID[candidate_reagent]
                                    )
                                    chosen_component = candidate_reagent
                            # Check if the number of wells in common is higher than the currently recorded number
                            elif (
                                len(
                                    current_reagent_well_set.intersection(
                                        Wells_By_Reagent_ID[candidate_reagent]
                                    )
                                )
                                > len(common_wells)
                                and not preferential_selected
                            ):
                                # If so, store this as the current best option (may get overwritten later)
                                common_wells = current_reagent_well_set.intersection(
                                    Wells_By_Reagent_ID[candidate_reagent]
                                )
                                chosen_component = candidate_reagent
                        chosen_components = [
                            chosen_component
                        ]  # make a list to make downstreaming processing easier

                        # Check if there were any identified reagents
                        if not chosen_components == [""]:
                            # print("chosen comps", chosen_components)
                            # If there were, then check if it satisifies all wells or not
                            if not len(current_reagent_well_set) == len(common_wells):
                                ## If the chosen component only applies to a subset of wells, activate the split
                                split_active = True
                                # Store the non-common wells in the split, along with the current reagents
                                split_tracker_wells = current_reagent_well_set.symmetric_difference(
                                    common_wells
                                )
                                split_tracker_reagents = mastermix_reagents.copy()
                                split.append(
                                    [split_tracker_wells.copy(), split_tracker_reagents.copy()]
                                )
                                # Remove split wells from the active mastermix_wells
                                current_reagent_well_set -= split_tracker_wells
                            else:
                                # If applies to all wells, then just pass
                                pass

                        else:
                            # print("chosen comps", chosen_components)
                            # if no remaining reagent has any common wells, then start looking at breaking apart current MMs
                            common_wells = set()
                            chosen_mastermix = None
                            for mastermix in Mastermixes:
                                # print(current_reagent_well_set, mastermix.wells)
                                if len(
                                    current_reagent_well_set.intersection(mastermix.wells)
                                ) > len(common_wells):
                                    # check that none of the current reagents are in the identified mastermix
                                    if (
                                        len(
                                            current_reagent_well_set.intersection(
                                                mastermix.reagents
                                            )
                                        )
                                        > 0
                                    ):
                                        continue
                                    # Make sure that the linked reagents aren't excluded combinations
                                    exclude = False
                                    for excluded_combo in Excluded_Combinations:
                                        for reagent_in_candidate_mm in mastermix.reagents:
                                            if (
                                                reagent_in_candidate_mm.split("_vol_")[0]
                                                in excluded_combo
                                                and reagent_id.split("_vol_")[0] in excluded_combo
                                            ):
                                                exclude = True
                                                break
                                    if exclude:
                                        continue
                                    common_wells = current_reagent_well_set.intersection(
                                        mastermix.wells
                                    )
                                    chosen_mastermix = mastermix
                            # Check if a mastermix was found
                            if chosen_mastermix:
                                # print("chosen mm", chosen_mastermix.name)
                                # Remove the common wells from the chosen mastermix as they will be re-assigned to the new mastermix
                                chosen_mastermix.wells -= common_wells
                                # If the mastermix is now not servicing any wells, delete it
                                if len(chosen_mastermix.wells) == 0:
                                    Mastermixes.remove(chosen_mastermix)
                                # Store the components from the chosen mastermix
                                chosen_components = chosen_mastermix.reagents.copy()
                                # If there were, then check if it satisifies all wells or not
                                if not len(current_reagent_well_set) == len(common_wells):
                                    # If the chosen component only applies to a subset of wells, activate the split
                                    split_active = True
                                    # Store the non-common wells in the split, along with the current reagents
                                    split_tracker_wells = (
                                        current_reagent_well_set.symmetric_difference(common_wells)
                                    )
                                    split_tracker_reagents = mastermix_reagents.copy()
                                    split.append(
                                        [split_tracker_wells.copy(), split_tracker_reagents.copy()]
                                    )
                                    # Remove split wells from the active mastermix_wells
                                    current_reagent_well_set = common_wells.copy()
                                else:
                                    # If applies to all wells, then just pass
                                    pass

                            else:
                                # If no mastermixes could be found to split, then raise an error
                                No_Solution = True
                                break

                        # update mastermix components
                        for chosen_component in chosen_components:
                            # print("Add comp", chosen_component)
                            mastermix_reagents.append(chosen_component)
                            mastermix_per_well += _get_vol_per_well(chosen_component)
                        # Check if the mm per well is above the threshold
                        if mastermix_per_well >= Min_Transfer_Volume:
                            # print("mm_per_well_sanity_check", mastermix_per_well, Min_Transfer_Volume)
                            # print("Reagent:", reagent_id)
                            # print("mm reag", mastermix_reagents)
                            # if yes, then mark the active well set as satisified and create a mastermix
                            mastermix_name = ":".join(mastermix_reagents)
                            mastermix_well_set = current_reagent_well_set.copy()  # active well set
                            Mastermixes.append(
                                Mastermix(
                                    mastermix_name, mastermix_reagents.copy(), mastermix_well_set
                                )
                            )
                            # Remove the used wells from the reagent IDs
                            for used_reagent in mastermix_reagents:
                                # print("Current {} well set: {}".format(used_reagent, Wells_By_Reagent_ID[used_reagent]))
                                Wells_By_Reagent_ID[used_reagent] -= mastermix_well_set
                                # print("Updated {} well set: {}".format(used_reagent, Wells_By_Reagent_ID[used_reagent]))
                            # print("---")
                            # Mark the current wells as satisified
                            current_reagent_well_set.clear()

                            if split_active:
                                ## Apply the most recent split as active
                                current_reagent_well_set = split[-1][0]
                                mastermix_reagents = split[-1][1]
                                mastermix_per_well = sum(
                                    [_get_vol_per_well(r) for r in mastermix_reagents]
                                )
                                split.remove(split[-1])
                                if len(split) == 0:
                                    split_active = False
                                else:
                                    pass
                        else:
                            # if mm/well is not above the threshold, then keep looping
                            continue

                    if No_Solution:
                        break

                else:
                    # If the reagent vol/well is already above the threshold, then just skip
                    continue

            if No_Solution:
                raise MastermixError("No solution could be found with these constraints.")
            else:
                Solved = True  # break out of the while loop with a solution

        # If a solution is not possible
        except MastermixError:
            # If this was the first attempt
            if First_Attempt:
                # Check if a seed was given
                if Seed:
                    raise MastermixError(
                        "No solution could be found with these constraints and a seed of {}".format(
                            Seed
                        )
                    )
                # If a seed wasn't given, then try shuffling the order of the reagents with different seeds
                else:
                    Seed = 0
                    First_Attempt = False
            else:
                # Iterate the seed number
                Seed += 1
                # To stop infinte loops
                if not First_Attempt:
                    if Seed > 1000:
                        raise MastermixError(
                            """
No solution could be found with these constraints. Try one of the following options:
> Decrease the minimum transfer volume (if specified)
> Decrease the stock concentration of highly concentrated source material
> Use a mastermix labware with a higher well capacity (especially when there are a lot of samples)
"""
                        )

    # Remove any empty mastermixes
    ## These are MMs which were generated at some point, but then were split out into other mastermixes and no longer service any wells

    for mm in Mastermixes:
        if sum([_get_vol_per_well(reagent) for reagent in mm.reagents]) < Min_Transfer_Volume:
            raise ValueError(mm.name)
        if len(mm.wells) == 0:
            Mastermixes.remove(mm)

    # Set up the mastermix layout with the generated content
    Mastermix_Layouts = [Mastermix_Layout]
    Mastermix_Layout_Index = 0

    for mastermix in Mastermixes:

        # Check if a new mastermix layout is needed
        if Mastermix_Layouts[Mastermix_Layout_Index].get_next_empty_well() is None:

            new_mastermix_layout_name = "{}_{}".format(
                Mastermix_Layout.name, Mastermix_Layout_Index + 1
            )
            Mastermix_Layouts.append(Mastermix_Layout.clone_format(new_mastermix_layout_name))
            Mastermix_Layout_Index += 1

        well = Mastermix_Layouts[Mastermix_Layout_Index].get_next_empty_well()
        Mastermix_Layouts[Mastermix_Layout_Index].add_well_label(well, mastermix.name)

        # Make sure that the transfer volume TO the mastermix is above the threshold for all reagents
        ## If not, additional extra reactions will be added to that mastermix
        extra_reactions = Extra_Reactions
        for reagent in mastermix.reagents:
            volume = (len(mastermix.wells) + extra_reactions) * (_get_vol_per_well(reagent))
            # Make sure that the transfer volume TO the mastermix is above the threshold
            if volume < Min_Transfer_Volume:
                extra_vol_needed = Min_Transfer_Volume - volume
                extra_reactions_needed = extra_vol_needed / _get_vol_per_well(reagent)
                extra_reactions += extra_reactions_needed
        # Add reagents to the mastermix layout
        for reagent in mastermix.reagents:
            Mastermix_Layouts[Mastermix_Layout_Index].add_content(
                Well=well,
                Reagent=reagent.split("_vol_")[0],
                Volume=(len(mastermix.wells) + extra_reactions) * (_get_vol_per_well(reagent)),
            )

    # Modify the destination layout to now take mastermixes as reagents
    for Destination_Layout, Dest_Index in zip(
        Destination_Layouts, range(0, len(Destination_Layouts))
    ):

        for destination_well in Destination_Layout.content.copy():
            destination_location = "{}_{}".format(Dest_Index, destination_well)
            for mastermix in Mastermixes:
                # Find if any mastermixes satisfy the destination well
                if destination_location in mastermix.wells:
                    # Get the reagents that the mastermix covers
                    reagents = mastermix.reagents
                    # Calculate the volume of mastermix needed for this well
                    mastermix_volume = sum([_get_vol_per_well(reagent) for reagent in reagents])
                    # Sanity check
                    if not mastermix_volume == sum(
                        [
                            Destination_Layout.get_volume_of_liquid_in_well(
                                reagent.split("_vol_")[0], destination_well
                            )
                            for reagent in reagents
                        ]
                    ):
                        # print("destination well content:", [reagent.split("_vol_")[0] for reagent in reagents], Destination_Layout.get_liquids_in_well(destination_well))
                        raise LabwareError(
                            "Mastermix Maker encountered an error with well {} mastermix {}: vol1 = {}, vol2 = {}".format(
                                destination_well,
                                mastermix.name,
                                mastermix_volume,
                                sum(
                                    [
                                        Destination_Layout.get_volume_of_liquid_in_well(
                                            reagent.split("_vol_")[0], destination_well
                                        )
                                        for reagent in reagents
                                    ]
                                ),
                            )
                        )

                    # Remove the reagents in the mastermix from the destination well
                    for reagent in reagents:
                        # print("remove from destination {}: {}".format(Dest_Index, reagent))
                        Destination_Layout.clear_liquid_in_well(
                            destination_well, reagent.split("_vol_")[0]
                        )

                    # Add mastermix to the destination well
                    Destination_Layout.add_content(
                        Well=destination_well, Reagent=mastermix.name, Volume=mastermix_volume
                    )

    return (Mastermixes, Seed, Destination_Layouts, Mastermix_Layouts)


def fmol_calculator(mass_ng, length_bp):
    return ((mass_ng * 1e-9) / ((length_bp * 617.96) + 36.04)) * 1e15

def serial_dilution_volumes(dilution_factors, total_volume):
    # Note that total volume is the amount the dilution will be made up to
    ## The total volume of all dilutions other than the final will be lower than this
    sample_volumes = []
    solution_volumes = []

    # This the the dilution factor of the source material for the first serial dilution
    ## This is always 1, as the initial sample is assumed to be undiluted
    source_dilution_factor = 1

    for df in dilution_factors:
        # Get the dilution factor of the current serial dilution being performed
        destination_dilution_factor = df

        # Calculate the volume, in uL, of sample and solution required for each dilution factor
        sample_volume = total_volume * (source_dilution_factor/destination_dilution_factor)
        solution_volume = total_volume - sample_volume

        # Store the volumes required for later use
        sample_volumes.append(sample_volume)
        solution_volumes.append(solution_volume)

        # Set the current dilution as the source for the next serial dilution
        source_dilution_factor = df

    return(sample_volumes, solution_volumes)

## Private ##
def _get_well_layout_index(well):
    return int(well.split("_")[0])


def _Lrange(L1, L2):  # Between L1 and L2 INCLUSIVE of L1 and L2
    L1 = ord(L1.upper())
    L2 = ord(L2.upper())
    for L in range(L1, L2 + 1):
        yield (chr(L))


def _Labware_Row_To_Index(row):
    return ord(row.upper()) - ord("A")


def _get_vol_per_well(Reagent_ID):
    return float(Reagent_ID.split("_vol_")[1])


class MastermixError(Exception):
    pass



import BiomationScripter as _BMS
import BiomationScripter.EchoProto as _EchoProto
import math
from typing import List, NewType, Dict, Tuple, Union
from decimal import Decimal


class Template(_EchoProto.EchoProto_Template):
    def __init__(self,
        Volume: float,
        Reactions: Tuple[str, str, str],
        Polymerase: str = None,
        Polymerase_Buffer: str = None,
        Polymerase_Buffer_Stock_Conc: float = None,
        Master_Mix: str = None,
        Master_Mix_Stock_Conc = None,
        Repeats: int = 1,
        DNA_Amounts = None,
        **kwargs
    ):

        super().__init__(**kwargs)

        ########################################
        # User defined aspects of the protocol #
        ########################################
        self.volume = Volume # uL
        self.repeats = Repeats
        self.reactions = Reactions
        self.polymerase = Polymerase
        self.buffer = Polymerase_Buffer
        self.master_mix = Master_Mix
        self.master_mix_stock_conc = Master_Mix_Stock_Conc
        self.buffer_stock_conc = Polymerase_Buffer_Stock_Conc

        if not self.master_mix:
            if not self.polymerase or not self.buffer:
                raise _BMS.BMSTemplateError("If mastermix is not specified, then the polymerase and buffer to use must be given.")

        ###########################
        # Default reagent amounts #
        ###########################
        self.__default_volume = 5

        self._dNTPs_amount = 0.1
        self._polymerase_amount = 0.05
        if self.master_mix:
            if not self.master_mix_stock_conc:
                raise _BMS.BMSTemplateError("If master mix is specified, a stock concentration (e.g. 2x) must also be given.")
            else:
                self._master_mix_amount = self.__default_volume / self.master_mix_stock_conc
        else:
            self._master_mix_amount = None
            self._buffer_amount = self.__default_volume / self.buffer_stock_conc

        ##################################
        # Default DNA and primer amounts #
        ##################################
        # Default DNA amounts in uL for 5 uL reactions, and 1 - 1000 ng/uL stock concentration
        self._default_dna_amount = 1 # uL
        if not DNA_Amounts:
            self.dna_amounts = [self._default_dna_amount * (self.volume/self.__default_volume)]
        else:
            self.dna_amounts = DNA_Amounts

        # Default primer amounts in uL for 5 uL reactions, and 10 μM stock concentration
        self.primer_amount = 0.25

        #########################
        # Default reagent names #
        #########################
        self.dNTPs = "dNTPs"
        self.water = "Water"

    def run(self):

        ###################################################
        # Calculate number of destination plates required #
        ###################################################
        # Determine final number of reactions, and hence number of destination wells required
        n_reactions = len(self.reactions) * self.repeats * len(self.dna_amounts)

        # Create the number of destination plates needed as layout objects
        extra_destination_layouts = _BMS.Create_Labware_Needed(
            Labware_Format = self.destination_plate_layouts[0],
            N_Wells_Needed = n_reactions,
            N_Wells_Available = len(self.destination_plate_layouts[0].get_available_wells()),
            Return_Original_Layout = False
        )

        # Add any extra destination plates needed
        for layout in extra_destination_layouts:
            self.add_destination_layout(layout)

        ######################################
        # Calculate reagent volumes required #
        ######################################
        # Volume factor is the difference between the default volume (5 uL) and the user-define final volume
        volume_factor = self.volume/self.__default_volume

        # Use the volume factor to calculate the reagent volumes
        primer_amount = self.primer_amount * volume_factor



        ########################################
        # Add liquids to destination layout(s) #
        ########################################
        # Counter to track which destination plate is in use
        destination_plate_index = 0
        # Counter to track which destination well is in use
        destination_well_index = 0

        # For every reaction
        for reaction in self.reactions:
            for dna_amount in self.dna_amounts:
                # Determine if a mastermix is being used
                if self.master_mix:
                    master_mix_amount = self._master_mix_amount * volume_factor
                    reagent_amount = master_mix_amount + dna_amount + (2 * primer_amount)
                else:
                    dNTPs_amount = self._dNTPs_amount * volume_factor
                    buffer_amount = self._buffer_amount * volume_factor
                    polymerase_amount = self._polymerase_amount * volume_factor
                    reagent_amount = float(Decimal(str(dNTPs_amount)) + Decimal(str(buffer_amount)) + Decimal(str(polymerase_amount)) + Decimal(str(dna_amount)) + (2 * Decimal(str(primer_amount))))

                water_amount = self.volume - reagent_amount

                for rep in range(0, self.repeats):
                    # Get the current destination plate and well
                    current_destination_plate = self.destination_plate_layouts[destination_plate_index]
                    current_destination_well = current_destination_plate.available_wells[destination_well_index]

                    # Add the reagents, water, and DNA to the destinaton plate
                    try:
                        current_destination_plate.add_content(current_destination_well, reaction[0], dna_amount)
                        current_destination_plate.add_content(current_destination_well, reaction[1], primer_amount)
                        current_destination_plate.add_content(current_destination_well, reaction[2], primer_amount)
                        current_destination_plate.add_content(current_destination_well, self.water, water_amount)

                        if self.master_mix:
                            current_destination_plate.add_content(current_destination_well, self.master_mix, master_mix_amount)
                        else:
                            current_destination_plate.add_content(current_destination_well, self.dNTPs, dNTPs_amount)
                            current_destination_plate.add_content(current_destination_well, self.buffer, buffer_amount)
                            current_destination_plate.add_content(current_destination_well, self.polymerase, polymerase_amount)

                        # Label the well
                        current_destination_plate.add_well_label(current_destination_well, "{}-{}-{}-DNA_Vol({})-{}".format(reaction[0], reaction[1], reaction[2], dna_amount, rep))
                    # Raise a more relevant error message if NegativeVolumeError occurs
                    except _BMS.NegativeVolumeError:
                        raise _BMS.NegativeVolumeError("This reaction is above the reaction volume: {}".format(reaction))

                    # Iterate to the next destination well
                    destination_well_index += 1
                    # Check if the current destinaton plate is full
                    if destination_well_index - 1 == len(current_destination_plate.available_wells):
                        # If so, iterate to the first well of the next destination plate
                        destination_plate_index += 1
                        destination_well_index = 0

        # Sanity check
        for dest in self.destination_plate_layouts:
            for well in dest.get_occupied_wells():
                reagents = dest.get_liquids_in_well(well)
                if not sum([dest.get_volume_of_liquid_in_well(reagent, well) for reagent in reagents]) == self.volume:
                    raise _BMS.BMSTemplateError("Internal Calculation Error. Please raise as an issue on the GitHub.")

        # Generate the picklists
        self.create_picklists()
